import threading

from PySide6.QtCore import QEvent, QCoreApplication

from . import GlobalInfo


class ExecuteCodeEvent(QEvent):
    def __init__(self, callable, args=None, kwargs=None):
        super().__init__(QEvent.User)
        self.callable = callable
        self.args = args
        self.kwargs = kwargs
        self.event = threading.Event()
        self.result = None
        self.exception = None

    def execute(self):
        if self.kwargs is None:
            if self.args is None:
                return self.callable()
            else:
                return self.callable(*self.args)
        else:
            if self.args is None:
                return self.callable(**self.kwargs)
            else:
                return self.callable(*self.args, **self.kwargs)


class GUIObjProxy:
    """
    Derived from http://code.activestate.com/recipes/496741-object-proxying/
    """
    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # proxying (special cases)
    #
    def __getattribute__(self, name):
        result = gui_thread_schedule(lambda: getattr(object.__getattribute__(self, "_obj"), name))
        if result is None:
            return result
        if type(result) in [int, float, str, bool]:
            return result
        return GUIObjProxy(result)

    def __delattr__(self, name):
        gui_thread_schedule(lambda: delattr(object.__getattribute__(self, "_obj"), name))

    def __setattr__(self, name, value):
        gui_thread_schedule(lambda: setattr(object.__getattribute__(self, "_obj"), name, value))

    def __nonzero__(self):
        return gui_thread_schedule(lambda: bool(object.__getattribute__(self, "_obj")))

    def __str__(self):
        return gui_thread_schedule(lambda: str(object.__getattribute__(self, "_obj")))

    def __repr__(self):
        return gui_thread_schedule(lambda: repr(object.__getattribute__(self, "_obj")))

    #
    # factories
    #
    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__',
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__',
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__',
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__',
        '__repr__', '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__',
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__',
        '__truediv__', '__xor__', 'next',
    ]

    @classmethod
    def _create_class_proxy(cls, theclass):
        """
        Creates a proxy for the given class.
        """

        def make_method(name):
            def method(self, *args, **kw):
                return gui_thread_schedule(lambda: getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw))

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)

    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        theclass.__init__(ins, obj, *args, **kwargs)
        return ins


def is_gui_thread():
    return threading.get_ident() == GlobalInfo.gui_thread or GlobalInfo.gui_thread is None


def gui_thread_schedule(callable, args=None, timeout=None):
    if is_gui_thread():
        if args is None:
            return callable()
        else:
            return callable(*args)

    event = ExecuteCodeEvent(callable, args)

    if GlobalInfo.is_test:
        GlobalInfo.add_event_during_test(event)
    else:
        try:
            QCoreApplication.postEvent(GlobalInfo.main_window, event)
        except RuntimeError:
            # the application is exiting and the main window has been destroyed. just let it go
            return None

    event.event.wait(timeout=timeout)  # TODO: unsafe. to be fixed later.
    if not event.event.is_set():
        # it timed out without getting scheduled to execute...
        return None

    if event.exception is not None:
        raise event.exception

    return event.result


def gui_thread_schedule_async(callable, args=None, kwargs=None):
    if is_gui_thread():
        if kwargs is None:
            if args is None:
                callable()
            else:
                callable(*args)
        else:
            if args is None:
                callable(**kwargs)
            else:
                callable(*args, **kwargs)
        return

    event = ExecuteCodeEvent(callable, args=args, kwargs=kwargs)

    if GlobalInfo.is_test:
        GlobalInfo.add_event_during_test(event)
    else:
        try:
            QCoreApplication.postEvent(GlobalInfo.main_window, event)
        except RuntimeError:
            # the application is exiting and the main window has been destroyed. just let it go
            pass




class ConfigurationEntry:
    """
    Describes a configuration entry in angr management.
    """

    __slots__ = ('name', 'type_', 'value', 'default_value')

    def __init__(self, name, type_, value, default_value=None):
        self.name = name
        self.type_ = type_
        self.value = value
        self.default_value = default_value

    def copy(self):
        return ConfigurationEntry(self.name, self.type_, self.value, default_value=self.default_value)



import asyncio
import logging
import threading
import os
from time import sleep
from getmac import get_mac_address as gma
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from angrmanagement.config import Conf
import angrmanagement.ui.views as Views
from ..base_plugin import BasePlugin


l = logging.getLogger(__name__)
l.setLevel('INFO')


try:
    from slacrs import Slacrs
    from slacrs.model import HumanActivity, HumanActivityEnum
except ImportError as ex:
    Slacrs = None  # type: Optional[type]


class LogHumanActivitiesPlugin(BasePlugin):
    """
    Log human activities
    """
    def __init__(self, *args, **kwargs):
        if not Slacrs:
            raise Exception("Skipping LogHumanActivities Plugin. Please install Slacrs.")
        super().__init__(*args, **kwargs)
        self._init_logger()
        self.session = None
        self.project_name = None
        self.project_md5 = None
        self._log_list = list()
        self.user = gma()
        self.active = True
        self.slacrs_thread = None

    def _init_logger(self): # pylint:disable=no-self-use
        user_dir = os.path.expanduser('~')
        log_dir = os.path.join(user_dir, "am-logging")
        os.makedirs(log_dir, exist_ok=True)
        log_file = os.path.join(log_dir, 'human_activities.log')
        fh = logging.FileHandler(log_file)
        fh.setLevel('INFO')
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        l.addHandler(fh)

    def on_workspace_initialized(self, workspace):
        self.slacrs_thread = threading.Thread(target=self._commit_logs)
        self.slacrs_thread.setDaemon(True)
        self.slacrs_thread.start()

    def handle_stack_var_renamed(self, func, offset, old_name, new_name):
        """
        Log a user's activity of variable renaming.
        """
        variable_rename = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.VariableRename,
            function=func._name,
            old_name=old_name,
            new_name=new_name,
            created_by=self.user,
        )
        self._log_list.append(variable_rename)
        l.debug("Add variable rename sesssion to slacrs")

    def handle_function_renamed(self, func, old_name, new_name):
        """
        Log a user's activity of function renaming.
        """
        function_rename = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.FunctionRename,
            addr=func.addr,
            old_name=old_name,
            new_name=new_name,
            created_by=self.user,
        )
        self._log_list.append(function_rename)
        l.debug("Add function rename sesssion to slacrs, project name %s, old_name %s, new_name %s",
                self.project_name, old_name, new_name)

    def handle_click_block(self, qblock, event):
        block_click = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.ClickBlock,
            addr=qblock.addr,
            created_by=self.user,
        )
        self._log_list.append(block_click)
        l.debug("Block %x is clicked", qblock.addr)
        return False

    def handle_click_insn(self, qinsn, event):
        insn_click = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.ClickInsn,
            addr=qinsn.addr,
            created_by=self.user,
        )
        self._log_list.append(insn_click)
        l.debug("Instruction %x is clicked", qinsn.addr)
        return False

    def handle_raise_view(self, view):
        view_name = view.__class__.__name__
        func = self._get_function_from_view(view)
        if func is not None and not func.am_none:
            func_name = func._name
            addr = func.addr
        else:
            func_name = None
            addr = None
        raise_view = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.RaiseView,
            view=view_name,
            created_by=self.user,
            function=func_name,
            addr=addr
        )
        self._log_list.append(raise_view)
        l.debug("View %s is raised with function %s", view_name, func_name)

    def handle_comment_changed(self, address, old_cmt, new_cmt, created: bool, decomp: bool):
        """
        Log a user's activity of changing comment
        @param new: T if a new comment. We don't log it in slacrs.
        @param comp: T if comment is in decompiler view
        """
        comment_change = HumanActivity(
            project=self.project_name,
            project_md5=self.project_md5,
            category=HumanActivityEnum.CommentChanged,
            addr=address,
            cmt=new_cmt,
            decomp=decomp,
            created_by=self.user,
        )
        self._log_list.append(comment_change)
        l.debug("Comment is added at %x", address)
        return False

    def handle_project_initialization(self):
        """
        Set project name
        """
        if self.workspace.main_instance.img_name is not None:
            self.project_name = self.workspace.main_instance.img_name
        else:
            filename = self.workspace.main_instance.project.filename
            self.project_name = filename
            self.project_md5 = self.workspace.main_instance.project.loader.main_object.md5.hex()
            l.debug("Set project md5 to %s", self.project_md5)
        l.debug("Set project name to %s", self.project_name)

    @staticmethod
    def _get_function_from_view(view):
        if isinstance(view, Views.DisassemblyView):
            return view._current_function
        if isinstance(view, (Views.CodeView, Views.ProximityView)):
            return view.function
        return None

    def _commit_logs(self):
        l.debug("database: %s", Conf.checrs_backend_str)
        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
        while self.active:
            try:
                sleep(3)
                connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
                if connector is None:
                    # chess connector does not exist
                    return None
                slacrs_instance = connector.slacrs_instance()
                if slacrs_instance is None:
                    # slacrs does not exist. continue
                    continue
                self.session = slacrs_instance.session()
                with self.session.no_autoflush:
                    while len(self._log_list) > 0:
                        log = self._log_list.pop()
                        self.session.add(log)
                    self.session.commit()
                self.session.close()
            except Exception:  # pylint:disable=broad-except
                pass

    def teardown(self):
        self.active = False



import logging
from typing import Optional, Union

from PySide6.QtGui import QFont, QStandardItemModel, QStandardItem
from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QHeaderView, QVBoxLayout, QTreeWidget, QTreeView, QLabel
from angr.knowledge_plugins import Function

try:
    from bintrace import TraceEvent
except ImportError as e:
    TraceEvent = 'TraceEvent'

from ...logic.debugger.bintrace import BintraceDebugger
from ...logic.debugger import DebuggerWatcher
from ...config import Conf
from .view import BaseView


_l = logging.getLogger(name=__name__)


class CallTreeModel(QStandardItemModel):
    """
    Model for the call tree.
    """
    Headers = ['Function']

    def hasChildren(self, index):
        item: Optional['CallTreeItem'] = self.itemFromIndex(index)
        if isinstance(item, CallTreeItem):
            return item.expandable
        return super().hasChildren(index)

    def headerData(self, section, orientation, role):  # pylint:disable=unused-argument
        if role != Qt.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None


class CallTreeItem(QStandardItem):
    """
    Item in call tree representing a function.
    """

    def __init__(self, function, event):
        name = hex(function) if isinstance(function, int) else function.name
        super().__init__(name)
        self.function: Union[int, Function] = function
        self.event: TraceEvent = event
        self.populated: bool = False
        self.expandable: bool = True


class CallExplorerView(BaseView):
    """
    Call Explorer view.
    """

    def __init__(self, instance, default_docking_position, *args, **kwargs):
        super().__init__('call_explorer', instance, default_docking_position, *args, **kwargs)

        self._last_updated_func: Optional[Union[int, Function]] = None
        self._inhibit_update: bool = False

        self.base_caption = 'Call Explorer'
        self._tree: Optional[QTreeWidget] = None
        self._init_widgets()
        self.reload()

        self.width_hint = 500
        self.height_hint = 400
        self.updateGeometry()

        self._dbg_manager = instance.debugger_mgr
        self._dbg_watcher = DebuggerWatcher(self._on_debugger_state_updated, self._dbg_manager.debugger)
        self._on_debugger_state_updated()

    @staticmethod
    def minimumSizeHint(*args, **kwargs):  # pylint:disable=unused-argument
        return QSize(200, 200)

    def _init_widgets(self):
        vlayout = QVBoxLayout()
        self._top_level_function_level = QLabel()
        self._reset_function_label()
        vlayout.addWidget(self._top_level_function_level)
        self._tree = QTreeView(self)
        self._model = CallTreeModel(self._tree)
        self._tree.setModel(self._model)
        self._tree.setFont(QFont(Conf.disasm_font))
        header = self._tree.header()
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        self._tree.expanded.connect(self._on_item_expanded)
        self._tree.clicked.connect(self._on_item_clicked)
        self._tree.doubleClicked.connect(self._on_item_double_clicked)
        vlayout.addWidget(self._tree)
        self.setLayout(vlayout)

    #
    # Events
    #

    def closeEvent(self, event):
        self._dbg_watcher.shutdown()
        super().closeEvent(event)

    def _on_item_clicked(self, index):
        """
        Highlights the corresponding call site.
        """
        item = self._model.itemFromIndex(index)

        # Do not try to update on a single click. Allow user to browse through the call tree
        original_inhibit = self._inhibit_update
        self._inhibit_update = True

        # Replay up to just before call
        dbg = self.instance.debugger_mgr.debugger
        dbg.replay_to_event(dbg._btrace.get_prev_exec_event(item.event, vcpu=dbg._trace_dbg.vcpu))

        self._inhibit_update = original_inhibit

    def _on_item_double_clicked(self, index):
        """
        Navigates into the call.
        """
        item = self._model.itemFromIndex(index)
        # Replay after the jump, jumping into the called function
        # FIXME: Doesn't consider proper selected debugger, assumes bintrace
        dbg = self.instance.debugger_mgr.debugger
        dbg.replay_to_event(dbg._btrace.get_next_exec_event(item.event, vcpu=dbg._trace_dbg.vcpu))

    def _on_item_expanded(self, index):
        """
        Descend into call tree for this node.
        """
        expanding_item = self._model.itemFromIndex(index)
        if not expanding_item.populated:
            dbg = self.instance.debugger_mgr.debugger
            if dbg.am_none:
                return
            called = dbg.get_called_functions(expanding_item.event)
            for func_or_addr, event in called:
                expanding_item.appendRow(CallTreeItem(func_or_addr, event))
            expanding_item.expandable = len(called) > 0
            expanding_item.populated = True

    def _on_debugger_state_updated(self):
        """
        Update current call state.
        """
        if self._inhibit_update:
            return

        dbg = self._dbg_watcher.debugger
        if isinstance(dbg.am_obj, BintraceDebugger):
            func = dbg.get_current_function()
            if func is not None:
                func = func[0]
        else:
            func = None

        if func is self._last_updated_func:
            return

        self._model.clear()
        self._last_updated_func = func

        if func is not None and isinstance(dbg.am_obj, BintraceDebugger):
            self._top_level_function_level.setText(f'Current function: {func.name}')
            for func, event in dbg.get_called_functions():
                self._model.appendRow(CallTreeItem(func, event))
        else:
            self._reset_function_label()

    def _reset_function_label(self):
        self._top_level_function_level.setText('Current function: Unknown')



import sys

from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QPalette

from ..widgets.qccode_highlighter import reset_formats
from ...data.object_container import ObjectContainer
from ...config import Conf
from ...logic import GlobalInfo


def repr_color(color):
    r,g,b,a = color.getRgb()
    return f'rgba({r},{g},{b},{a})'

#
# FIXME: Need to propagate CSS updates to more objects.
#

class CSS:

    global_css = ObjectContainer('', 'Global CSS')

    @staticmethod
    def rebuild():
        CSS.global_css.am_obj = """
QLabel[class=insn] {
    font: 10pt courier new;
    color: #000080;
}

QLabel[class=reg_viewer_label] {
    font: 10pt courier new;
    background-color: #ffffff;
}

QLabel[class=ast_viewer_size] {
    font: 10pt courier new;
}

QLabel[class=ast_viewer_ast_concrete] {
    font: 10pt courier new;
    color: blue;
}

QLabel[class=ast_viewer_ast_symbolic] {
    font: 10pt courier new;
    color: green;
}

QLabel[class=memory_viewer_address] {
    font: 10pt courier new;
}

QLabel[class=status_valid] {
    color: green;
}

QLabel[class=status_invalid] {
    color: red;
}

QFrame[class=insn_selected] {
    font: 10pt courier new;
    color: #000000;
    background-color: #efbfba;
}

QBlock {
    border: 1px solid black;
}

QBlockLabel {
    color: #0000ff;
}

QLabel[class=insn_addr] {
    font: 10pt courier new;
    color: black;
}

QLabel[class=insn_string] {
    font: 10pt courier new;
    color: gray;
    font-weight: bold;
}

QDockWidget::title {
    """ + f"background: {repr_color(Conf.palette_mid)};" + """
    border: 1px solid gray;
    padding: 0px 0px 0px 5px;
    margin: 0px 0px 2px 0px;
}

QPlainTextEdit, QTextEdit {
    """ + f"background-color: {repr_color(Conf.palette_base)};" + """
}

QTableView {
    """ + f"background-color: {repr_color(Conf.palette_base)};" + """
}
"""
        CSS.global_css.am_event()


def refresh_theme():
    app = QApplication.instance()

    # determine the default application style according to the OS
    if sys.platform == "win32":
        if Conf.theme_name == "Light":
            app_style = None
        else:
            app_style = "Fusion"
    elif sys.platform == "darwin":
        if Conf.theme_name == "Light":
            app_style = None
        else:
            app_style = "Fusion"
    elif sys.platform == "linux":
        app_style = "Fusion"
    else:
        app_style = "Fusion"

    if app_style:
        app.setStyle(app_style)

    palette = QPalette()
    palette.setColor(QPalette.Window,          Conf.palette_window)
    palette.setColor(QPalette.WindowText,      Conf.palette_windowtext)
    palette.setColor(QPalette.Base,            Conf.palette_base)
    palette.setColor(QPalette.AlternateBase,   Conf.palette_alternatebase)
    palette.setColor(QPalette.ToolTipBase,     Conf.palette_tooltipbase)
    palette.setColor(QPalette.ToolTipText,     Conf.palette_tooltiptext)
    palette.setColor(QPalette.Text,            Conf.palette_text)
    palette.setColor(QPalette.Button,          Conf.palette_button)
    palette.setColor(QPalette.ButtonText,      Conf.palette_buttontext)
    palette.setColor(QPalette.BrightText,      Conf.palette_brighttext)
    palette.setColor(QPalette.Highlight,       Conf.palette_highlight)
    palette.setColor(QPalette.HighlightedText, Conf.palette_highlightedtext)
    palette.setColor(QPalette.Light,           Conf.palette_light)
    palette.setColor(QPalette.Midlight,        Conf.palette_midlight)
    palette.setColor(QPalette.Dark,            Conf.palette_dark)
    palette.setColor(QPalette.Mid,             Conf.palette_mid)
    palette.setColor(QPalette.Shadow,          Conf.palette_shadow)
    palette.setColor(QPalette.Link,            Conf.palette_link)
    palette.setColor(QPalette.LinkVisited,     Conf.palette_linkvisited)
    palette.setColor(QPalette.Disabled, QPalette.Text,       Conf.palette_disabled_text)
    palette.setColor(QPalette.Disabled, QPalette.ButtonText, Conf.palette_disabled_buttontext)
    palette.setColor(QPalette.Disabled, QPalette.WindowText, Conf.palette_disabled_windowtext)
    app.setPalette(palette)
    CSS.rebuild()
    app.setStyleSheet(CSS.global_css.am_obj)

    reset_formats()

    if GlobalInfo.main_window is not None:
        for codeview in GlobalInfo.main_window.workspace.view_manager.views_by_category['pseudocode']:
            codeview.codegen.am_event(already_regenerated=True)

            if codeview._textedit is not None:
                for panel in codeview._textedit.panels:
                    panel.setPalette(palette)



from typing import Optional, TYPE_CHECKING

from PySide6.QtCore import Qt, QEvent
from PySide6.QtGui import QTextCharFormat, QKeySequence, QAction
from PySide6.QtWidgets import QMenu, QInputDialog, QLineEdit, QApplication

from pyqodeng.core import api
from pyqodeng.core import modes
from pyqodeng.core import panels

from ailment.statement import Store, Assignment
from ailment.expression import Load, Op, UnaryOp, BinaryOp
from angr.sim_type import SimType
from angr.sim_variable import SimVariable, SimTemporaryVariable
from angr.analyses.decompiler.structured_codegen.c import CBinaryOp, CVariable, CFunctionCall, CFunction, \
    CStructField, CIndexedVariable, CVariableField, CUnaryOp, CConstant, CExpression
from angr.analyses.decompiler.optimization_passes.expr_op_swapper import OpDescriptor

from ..documents.qcodedocument import QCodeDocument
from ..dialogs.rename_node import RenameNode
from ..dialogs.retype_node import RetypeNode
from ..widgets.qccode_highlighter import QCCodeHighlighter
from ..menus.menu import Menu

if TYPE_CHECKING:
    from PySide6.QtGui import QTextDocument
    from ..views.code_view import CodeView


class ColorSchemeIDA(api.ColorScheme):
    """
    An IDA-like color scheme.
    """
    def __init__(self):
        super().__init__('default')

        # override existing formats
        function_format = QTextCharFormat()
        function_format.setForeground(self._get_brush("0000ff"))
        self.formats['function'] = function_format


class QCCodeEdit(api.CodeEdit):
    """
    A subclass of pyqodeng's CodeEdit, specialized to handle the kinds of textual interaction expected of the pseudocode
    view. You will typically interact with this class as code_view.textedit.
    """
    def __init__(self, code_view):
        super().__init__(create_default_actions=True)

        self._code_view: 'CodeView' = code_view

        self.panels.append(panels.LineNumberPanel())
        self.panels.append(panels.FoldingPanel())

        self.modes.append(modes.SymbolMatcherMode())

        self.setTabChangesFocus(False)
        self.setReadOnly(True)

        self.constant_actions = [ ]
        self.operator_actions = [ ]
        self.variable_actions = [ ]
        self.selected_actions = [ ]
        self.call_actions = [ ]
        self.default_actions = [ ]
        self.function_name_actions = []
        self.action_rename_node = None
        self._selected_node = None

        self._initialize_context_menus()

        # but we don't need some of the actions
        self.remove_action(self.action_undo)
        self.remove_action(self.action_redo)
        self.remove_action(self.action_cut)
        self.remove_action(self.action_paste)
        self.remove_action(self.action_duplicate_line)
        self.remove_action(self.action_swap_line_up)
        self.remove_action(self.action_swap_line_down)

    def node_under_cursor(self):
        doc: 'QTextDocument' = self.document()
        if not isinstance(doc, QCodeDocument):
            # this is not the qcodedocument that the decompiler generates. this means the pseudocode view is empty
            return None

        # determine the current status
        cursor = self.textCursor()
        pos = cursor.position()
        current_node = doc.get_node_at_position(pos)
        if current_node is not None:
            return current_node
        else:
            # nothing is under the cursor
            return None

    def get_context_menu(self):
        if self.document() is None:
            return QMenu()

        # TODO: Anything in sel?

        # Get the highlighted item
        under_cursor = self.node_under_cursor()

        mnu = QMenu()
        self._selected_node = None
        if isinstance(under_cursor, CConstant):
            self._selected_node = under_cursor
            mnu.addActions(self.constant_actions)
        if isinstance(under_cursor, (CBinaryOp, CUnaryOp)):
            # operator in selection
            self._selected_node = under_cursor
            mnu.addActions(self.operator_actions)
        if isinstance(under_cursor, CFunctionCall) \
                and "vex_block_addr" in under_cursor.tags \
                and "ins_addr" in under_cursor.tags:
            # function call in selection
            self._selected_node = under_cursor
            mnu.addActions(self.call_actions)
        if isinstance(under_cursor, (CVariable, CIndexedVariable, CVariableField, CStructField)):
            # variable in selection
            self._selected_node = under_cursor
            mnu.addActions(self.variable_actions)
        if isinstance(under_cursor, CFunction):
            # decompiled function name in selection
            self._selected_node = under_cursor
            mnu.addActions(self.function_name_actions)
            for entry in self.workspace.plugins.build_context_menu_functions(
                    [self.instance.kb.functions[under_cursor.name]]):
                Menu.translate_element(mnu, entry)
        else:
            mnu.addActions(self.default_actions)

        for entry in self.workspace.plugins.build_context_menu_node(under_cursor):
            Menu.translate_element(mnu, entry)

        return mnu

    @property
    def workspace(self):
        return self._code_view.instance.workspace if self._code_view is not None else None

    @property
    def instance(self):
        return self._code_view.instance if self._code_view is not None else None

    def event(self, event):
        """
        Reimplemented to capture the Tab key pressed/released event.

        :param event:
        :return:
        """

        if event.type() == QEvent.ShortcutOverride and event.key() == Qt.Key_Tab:
            event.accept()
            return True
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            return self.keyPressEvent(event)

        return super().event(event)

    def get_closest_insaddr(self, node, expr=None) -> Optional[int]:
        addr = (getattr(node, 'tags', None) or {}).get('ins_addr', None)
        if addr is None:
            if expr:
                addr = self.get_src_to_inst()
            else:
                pos = self.textCursor().position()
                while self.document().characterAt(pos) not in ('\n', '\u2029') and \
                        pos < self.document().characterCount():  # qt WHAT are you doing
                    pos += 1
                node = self.document().get_stmt_node_at_position(pos)
                addr = (getattr(node, 'tags', None) or {}).get('ins_addr', None)

        return addr

    def get_src_to_inst(self) -> int:
        """
        Uses the current cursor position, which is in a code view, and gets the
        corresponding instruction address that is associated to the code.
        Returns the start of the function if unable to calculate.

        :return: int (address of inst)
        """

        # get the Qt document
        doc: 'QCodeDocument' = self.document()

        # get the current position of the cursor
        cursor = self.textCursor()
        pos = cursor.position()

        # get the node at the associated cursor position
        current_node = doc.get_stmt_node_at_position(pos)

        if current_node is not None and hasattr(current_node, 'tags') and \
                current_node.tags is not None and 'ins_addr' in current_node.tags:
            asm_ins_addr = current_node.tags['ins_addr']

        else:
            # the top of the function decompiled
            asm_ins_addr = self._code_view.function.addr

        return asm_ins_addr

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        xkey = key
        if isinstance(xkey, int):
            xkey = Qt.Key(xkey)
        if modifiers & Qt.ShiftModifier:
            xkey |= Qt.SHIFT
        if modifiers & Qt.ControlModifier:
            xkey |= Qt.CTRL
        if modifiers & Qt.AltModifier:
            xkey |= Qt.ALT
        if modifiers & Qt.MetaModifier:
            xkey |= Qt.META
        sequence = QKeySequence(xkey)
        mnu = self.get_context_menu()
        for item in mnu.actions():
            if item.shortcut().matches(sequence) == QKeySequence.SequenceMatch.ExactMatch:
                item.activate(QAction.ActionEvent.Trigger)
                return True

        if key in (Qt.Key_Slash, Qt.Key_Question):
            self.comment(expr=event.modifiers() & Qt.ShiftModifier == Qt.ShiftModifier)
            return True
        if key == Qt.Key_Minus and QApplication.keyboardModifiers() & Qt.CTRL != 0:
            self.zoomOut()
        if key == Qt.Key_Equal and QApplication.keyboardModifiers() & Qt.CTRL != 0:
            self.zoomIn()

        if self._code_view.keyPressEvent(event):
            return True

        saved_mode = self.textInteractionFlags()
        if key in (Qt.Key_Left, Qt.Key_Right, Qt.Key_Up, Qt.Key_Down,
                   Qt.Key_PageDown, Qt.Key_PageUp, Qt.Key_Home, Qt.Key_End):
            self.setTextInteractionFlags(saved_mode | Qt.TextEditable)
        result = super().keyPressEvent(event)
        self.setTextInteractionFlags(saved_mode)
        return result

    def paintEvent(self, e):
        saved_mode = self.textInteractionFlags()
        self.setTextInteractionFlags(saved_mode | Qt.TextEditable)
        super().paintEvent(e)
        self.setTextInteractionFlags(saved_mode)

    def setDocument(self, document):
        super().setDocument(document)

        self.modes.append(QCCodeHighlighter(self.document(), color_scheme=ColorSchemeIDA()))
        self.syntax_highlighter.fold_detector = api.CharBasedFoldDetector()

    #
    # Actions
    #

    def rename_node(self, *args, node=None):  # pylint: disable=unused-argument
        n = node if node is not None else self._selected_node
        if not isinstance(n, (CVariable, CFunction, CFunctionCall, CStructField, SimType)):
            return
        if isinstance(n, CVariable) and isinstance(n.variable, SimTemporaryVariable):
            # unsupported right now..
            return
        dialog = RenameNode(code_view=self._code_view, node=n)
        dialog.exec_()

    def retype_node(self, *args, node=None, node_type=None):  # pylint: disable=unused-argument
        if node is None:
            node = self._selected_node
        if not isinstance(node, (CVariable, CFunction, CFunctionCall, CStructField)):
            return
        if isinstance(node, CVariable) and isinstance(node.variable, SimTemporaryVariable):
            # unsupported right now..
            return
        dialog = RetypeNode(self.instance, code_view=self._code_view, node=node,
                            node_type=node_type)
        dialog.exec_()

        new_node_type = dialog.new_type
        if new_node_type is not None:
            if self._code_view is not None and node is not None:
                # need workspace for altering callbacks of changes
                variable_kb = self._code_view.codegen._variable_kb
                # specify the type
                new_node_type = new_node_type.with_arch(self.instance.project.arch)
                variable_kb.variables[self._code_view.function.addr].set_variable_type(
                    node.variable,
                    new_node_type,
                    all_unified=True,
                    mark_manual=True,
                )

                self._code_view.codegen.am_event(event="retype_variable", node=node, variable=node.variable)

    def comment(self, expr=False, node=None):
        addr = self.get_closest_insaddr(node, expr=expr)
        if addr is None:
            return

        try:
            cdict = self._code_view.codegen.expr_comments if expr else self._code_view.codegen.stmt_comments
        except AttributeError:
            return
        text = cdict.get(addr, "")

        text, ok = QInputDialog.getText(
            self._code_view,
            "Expression Comment" if expr else "Statement Comment",
            "",
            QLineEdit.Normal,
            text)

        if ok:
            exists = addr in cdict
            if text:
                # callback
                self.workspace.plugins.handle_comment_changed(addr, "", text, not exists, True)
                cdict[addr] = text
            else:
                if exists:
                    # callback
                    self.workspace.plugins.handle_comment_changed(addr, cdict[addr], "", False, True)
                    del cdict[addr]

            self._code_view.codegen.am_event()

    def toggle_struct(self):
        node = self._selected_node
        if not isinstance(node, CVariable):
            return
        if isinstance(node.variable, SimVariable):
            ident = node.variable.ident
            if ident in self._code_view.vars_must_struct:
                self._code_view.vars_must_struct.remove(ident)
            else:
                self._code_view.vars_must_struct.add(ident)
            # decompile again
            self._code_view.decompile()

    def collapse_expr(self):
        if hasattr(self._selected_node, 'collapsed'):
            self._selected_node.collapsed = True
            self._code_view.codegen.am_event()

    def expand_expr(self):
        if hasattr(self._selected_node, 'collapsed'):
            self._selected_node.collapsed = False
            self._code_view.codegen.am_event()

    def hex_constant(self):
        if hasattr(self._selected_node, 'fmt_hex'):
            self._selected_node.fmt_hex ^= True
            self._code_view.codegen.am_event()

    def neg_constant(self):
        if hasattr(self._selected_node, 'fmt_neg'):
            self._selected_node.fmt_neg ^= True
            self._code_view.codegen.am_event()

    def convert_to_ite_expr(self):
        node = self._selected_node
        if not isinstance(node, CExpression):
            return
        ailexpr = self._code_view.codegen.cnode2ailexpr.get(node, None)
        if ailexpr is None:
            return

        # which statement?
        addr = self.get_closest_insaddr(node)
        if addr is None:
            return

        cache = self.instance.kb.structured_code[(self._code_view.function.addr, "pseudocode")]
        if cache.ite_exprs is None:
            cache.ite_exprs = set()
        cache.ite_exprs.add((addr, ailexpr))
        self._code_view.decompile(clear_prototype=False, regen_clinic=False)

    def swap_binop_operands(self):
        node = self._selected_node
        if not isinstance(node, CBinaryOp):
            return
        ailexpr = self._code_view.codegen.cnode2ailexpr.get(node, None)
        if ailexpr is None:
            return
        if not isinstance(ailexpr, BinaryOp):
            return

        op = ailexpr.op
        if op in {'CmpEQ', 'CmpNE'}:
            negated_op = op
        else:
            negated_op = BinaryOp.COMPARISON_NEGATION.get(op, None)
        if negated_op is None:
            return

        # which statement?
        addr = self.get_closest_insaddr(node)
        if addr is None:
            return

        cache = self.instance.kb.structured_code[(self._code_view.function.addr, "pseudocode")]
        if cache.binop_operators is None:
            cache.binop_operators = { }
        op_desc = OpDescriptor(ailexpr.vex_block_addr if hasattr(ailexpr, "vex_block_addr") else None,
                               ailexpr.vex_stmt_idx if hasattr(ailexpr, "vex_stmt_idx") else None,
                               addr,
                               op,
                               )
        cache.binop_operators[op_desc] = negated_op

        if negated_op != op:
            existing_op_desc = OpDescriptor(
                ailexpr.vex_block_addr if hasattr(ailexpr, "vex_block_addr") else None,
                ailexpr.vex_stmt_idx if hasattr(ailexpr, "vex_stmt_idx") else None,
                addr,
                negated_op,
            )
            if existing_op_desc in cache.binop_operators:
                del cache.binop_operators[existing_op_desc]
        self._code_view.decompile(clear_prototype=False, regen_clinic=False)

    def expr2armasm(self):

        def _assemble(expr, expr_addr) -> str:
            return converter.assemble(expr, self._code_view.function.addr, expr_addr)

        def _find_loads(expr) -> list:
            if isinstance(expr, Load):
                return [expr]
            elif isinstance(expr, Op):
                if isinstance(expr, UnaryOp):
                    return _find_loads(expr.operand)
                else:
                    loads = []
                    for operand in expr.operands:
                        loads += _find_loads(operand)
                    return loads
            else:
                return []

        node = self._selected_node
        # figure out where we are
        if not isinstance(node, (CVariable, CIndexedVariable, CVariableField, CStructField)):
            return

        doc: 'QCodeDocument' = self.document()
        cursor = self.textCursor()
        pos = cursor.position()
        current_node = doc.get_stmt_node_at_position(pos)

        if current_node is None:
            return

        ins_addr = current_node.tags.get("ins_addr", None)
        if ins_addr is None:
            return

        # traverse the stored clinic graph to find the AIL block
        cache = self.instance.kb.structured_code[(self._code_view.function.addr, "pseudocode")]
        the_node = None
        for node in cache.clinic.graph.nodes():
            if node.addr <= ins_addr < node.addr + node.original_size:
                the_node = node
                break

        if the_node is None:
            return

        converter = self.workspace.plugins.get_plugin_instance_by_name("AIL2ARM32")

        lst = [ ]
        for stmt in the_node.statements:
            if isinstance(stmt, Assignment):
                loads = _find_loads(stmt.src)
                for load in loads:
                    asm = _assemble(load, load.ins_addr)
                    lst.append((str(stmt), str(load), asm))
            elif isinstance(stmt, Store):
                loads = _find_loads(stmt.data)
                for load in loads:
                    asm = _assemble(load, load.ins_addr)
                    lst.append((str(stmt), str(load), asm))

        # format text
        text = f"The AIL block:\n{str(the_node)}\n\n"
        for stmt, expr, asm in lst:
            text += f"Statement: {stmt}\n"
            text += f"Expression: {expr}\n"
            text += f"Assembly:\n{asm}\n\n"

        converter.display_output(text)

    #
    # Private methods
    #

    @staticmethod
    def _separator():
        sep = QAction()
        sep.setSeparator(True)
        return sep

    def _initialize_context_menus(self):

        base_actions = [
            self._separator(),
            self.action_copy,
            self.action_select_all,
        ]

        self.action_rename_node = QAction('Rename variable', self)
        self.action_rename_node.triggered.connect(self.rename_node)
        self.action_rename_node.setShortcut(QKeySequence('N'))
        self.action_retype_node = QAction("Retype variable", self)
        self.action_retype_node.triggered.connect(self.retype_node)
        self.action_retype_node.setShortcut(QKeySequence('Y'))
        self.action_toggle_struct = QAction('Toggle &struct/array')
        self.action_toggle_struct.triggered.connect(self.toggle_struct)
        self.action_collapse_expr = QAction('Collapse expression', self)
        self.action_collapse_expr.triggered.connect(self.collapse_expr)
        self.action_expand_expr = QAction('Expand expression', self)
        self.action_expand_expr.triggered.connect(self.expand_expr)
        self.action_hex = QAction('Toggle hex', self)
        self.action_hex.triggered.connect(self.hex_constant)
        self.action_hex.setShortcut(QKeySequence('H'))
        self.action_neg = QAction('Toggle negative', self)
        self.action_neg.triggered.connect(self.neg_constant)
        self.action_neg.setShortcut(QKeySequence('_'))
        self.action_to_ite_expr = QAction("Create a ternary expression")
        self.action_to_ite_expr.triggered.connect(self.convert_to_ite_expr)
        self.action_swap_binop_operands = QAction("Swap operands")
        self.action_swap_binop_operands.triggered.connect(self.swap_binop_operands)

        expr_actions = [
            self.action_to_ite_expr,
            self.action_swap_binop_operands,
            self.action_collapse_expr,
            self.action_expand_expr,
        ]

        self.action_asmgen = QAction("Expression -> ARM THUMB assembly...")
        self.action_asmgen.triggered.connect(self.expr2armasm)

        self.variable_actions = [
            self.action_rename_node,
            self.action_retype_node,
            self.action_toggle_struct,
            self.action_asmgen,
        ]

        self.function_name_actions = [
            self.action_rename_node,
        ]

        self.constant_actions = [
            self.action_hex,
            self.action_neg,
        ]

        self.constant_actions += base_actions + expr_actions
        self.operator_actions += base_actions + expr_actions
        self.variable_actions += base_actions + expr_actions
        self.function_name_actions += base_actions
        self.call_actions += base_actions + expr_actions
        self.selected_actions += base_actions
        self.default_actions += base_actions



__version__ = "9.2.25"


try:
    # make sure qtpy (which is used in PyQodeNG.core) is using PySide6
    import os
    os.environ['QT_API'] = 'pyside6'
    import qtpy
except ImportError:
    # qtpy is not installed
    pass



import logging


class LogRecord:
    """
    Stores a log record.
    """

    __slots__ = ('level', 'timestamp', 'source', 'content', )

    def __init__(self, level, timestamp, source, content):
        self.level = level
        self.timestamp = timestamp
        self.source = source
        self.content = content


class LogDumpHandler(logging.Handler):
    """
    Dumps log messages.
    """

    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance = instance

    def emit(self, record: logging.LogRecord) -> None:
        log_record = LogRecord(record.levelno, record.created, record.name, self.format(record))
        self.instance.log.append(log_record)
        self.instance.log.am_event(log_record=log_record)



from .analysis_options import AnalysisOptionsDialog
from .load_binary import LoadBinary
from .load_plugins import LoadPlugins
from .load_docker_prompt import LoadDockerPrompt
from .preferences import Preferences
from .breakpoint import BreakpointDialog



import logging

from PySide6.QtWidgets import QWidget, QVBoxLayout, QGraphicsScene, QGraphicsView, QGraphicsItemGroup
from PySide6.QtWidgets import QTabWidget, QPushButton, QAbstractItemView
from PySide6.QtWidgets import  QMessageBox, QTableWidget, QTableWidgetItem
from PySide6.QtGui import QPen, QBrush, QLinearGradient, QColor, QPainter, QImage, QFont
from PySide6.QtCore import Qt, QPoint, QEvent


l = logging.getLogger(name=__name__)


class QTraceViewer(QWidget):
    """
    Load a basic block trace through json and visualize it in the disassembly
    Ref: https://github.com/angr/angr-management/pull/122
    """

    TAG_SPACING = 50
    LEGEND_X = -50
    LEGEND_Y = 0
    LEGEND_WIDTH = 10

    TRACE_FUNC_X = 0
    TRACE_FUNC_Y = 0
    TRACE_FUNC_WIDTH = 50
    TRACE_FUNC_MINHEIGHT = 1000

    TAB_HEADER_SIZE = 40
    MAX_WINDOW_SIZE = 500

    MARK_X = LEGEND_X
    MARK_WIDTH = TRACE_FUNC_X - LEGEND_X + TRACE_FUNC_WIDTH
    MARK_HEIGHT = 1

    def __init__(self, workspace, disasm_view, parent=None):
        super().__init__(parent=parent)
        self.workspace = workspace
        self.disasm_view = disasm_view

        self.mark = None
        self.legend = None
        self.legend_height = 0
        self.legend_img = None
        self.trace_func_unit_height = 0


        self.trace_func = None
        self.trace_id = None

        self.view = None
        self.traceView = None
        self.traceTab = None
        self.traceScene = None
        self.multiView = None
        self.listView = None
        self.mark = None
        self.curr_position = 0
        self._use_precise_position = False
        self._selected_traces = []

        self._init_widgets()

        self.trace.am_subscribe(self._on_set_trace)
        self.selected_ins.am_subscribe(self._on_select_ins)
        self.traceTab.installEventFilter(self)

    #
    # Forwarding properties
    #

    @property
    def trace(self):
        return self.workspace.main_instance.trace

    @property
    def multi_trace(self):
        return self.workspace.main_instance.multi_trace

    @property
    def selected_ins(self):
        return self.disasm_view.infodock.selected_insns

    def _init_widgets(self):
        self.view = QTabWidget() # QGraphicsView()

        self.traceTab = QWidget()
        tracelayout = QVBoxLayout()

        self.traceView = QGraphicsView()
        self.traceScene = QGraphicsScene()
        self.traceView.setScene(self.traceScene)

        self.listView = QTableWidget(0,2) # row, col
        self.listView.setHorizontalHeaderItem(0, QTableWidgetItem("Trace ID"))
        self.listView.setHorizontalHeaderItem(1, QTableWidgetItem("Input ID"))
        self.listView.setSelectionMode(QAbstractItemView.SingleSelection)
        self.listView.setSelectionBehavior(QAbstractItemView.SelectRows)
        # self.listView.horizontalHeader().setStretchLastSection(True)
        # self.listView.horizontalHeader().setSectionResizeModel(0, QHeaderView.Stretch)
        self.listView.cellClicked.connect(self._switch_current_trace)

        self.traceSeedButton = QPushButton("View Input Seed")
        self.traceSeedButton.clicked.connect(self._view_input_seed)

        tracelayout.addWidget(self.traceView)
        tracelayout.addWidget(self.listView)
        tracelayout.addWidget(self.traceSeedButton)
        self.traceTab.setLayout(tracelayout)

        self.multiView = QWidget()
        multiLayout = QVBoxLayout()
        self.multiTraceList = QTableWidget(0,2) # row, col
        self.multiTraceList.setSelectionMode(QAbstractItemView.MultiSelection)
        self.multiTraceList.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.multiTraceList.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.multiTraceList.setHorizontalHeaderItem(0, QTableWidgetItem("Trace ID"))
        self.multiTraceList.setHorizontalHeaderItem(1, QTableWidgetItem("Input ID"))
        self.selectMultiTrace = QPushButton("Refresh Heatmap")
        self.selectMultiTrace.clicked.connect(self._refresh_heatmap)
        multiLayout.addWidget(self.multiTraceList)
        multiLayout.addWidget(self.selectMultiTrace)
        self.multiView.setLayout(multiLayout)

        self.view.addTab(self.traceTab, "SingleTrace")
        self.view.addTab(self.multiView, "MultiTrace HeatMap")
        self.SINGLE_TRACE = 0
        self.MULTI_TRACE = 1

        self.view.currentChanged.connect(self._on_tab_change)

        self._reset()

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(self.view, Qt.AlignLeft)

        self.setLayout(layout)

    def _reset(self):
        self.traceScene.clear() #clear items
        self.listView.clearContents()
        self.multiTraceList.clearContents()
        self.mark = None

        self.legend = None
        self.legend_height = 0

        self.trace_func = QGraphicsItemGroup()
        self.trace_id = QGraphicsItemGroup()
        self.traceScene.addItem(self.trace_func)
        self.hide()

    def _view_input_seed(self):
        current_trace_stats = self.trace.am_obj
        input_id = current_trace_stats.input_id

        inputSeed = self.multi_trace.am_obj.get_input_seed_for_id(input_id)
        msgText = "%s" % inputSeed
        msgDetails = "Input for [%s]" % current_trace_stats.id
        msgbox = QMessageBox()
        msgbox.setWindowTitle("Seed Input")
        msgbox.setDetailedText(msgDetails)
        msgbox.setText(msgText)
        msgbox.setStandardButtons(QMessageBox.Ok)
        msgbox.exec()

    def _switch_current_trace(self, row):
        if self.listView.rowCount() <= 0:
            return

        current_trace = self.trace.am_obj.id
        new_trace = self.multiTraceList.item(row, 0).text()
        if current_trace == new_trace:
            return

        trace_stats = self.multi_trace.am_obj.get_trace_with_id(new_trace)
        if trace_stats:
            self.trace.am_obj = trace_stats
            self._on_set_trace()

    def _on_set_trace(self):
        self._reset()
        if self.trace.am_none or self.trace.count is None:
            return

        l.debug('minheight: %d, count: %d', self.TRACE_FUNC_MINHEIGHT,
                self.trace.count)
        if self.trace.count <= 0:
            l.warning("No valid addresses found in trace to show. Check base address offsets?")
            self.trace.am_obj = None
            self.trace.am_event()
            return
        if self.TRACE_FUNC_MINHEIGHT < self.trace.count * 15:
            self.trace_func_unit_height = 15
            show_func_tag = True
        else:
            self.trace_func_unit_height = self.TRACE_FUNC_MINHEIGHT / self.trace.count
            show_func_tag = True

        self.legend_height = int(self.trace.count * self.trace_func_unit_height)

        self._show_trace_func(show_func_tag)
        self._show_legend()
        self._show_trace_ids()
        self._set_mark_color()
        self._refresh_multi_list()

        boundingSize = self.traceScene.itemsBoundingRect().width()
        windowSize = boundingSize
        if boundingSize > self.MAX_WINDOW_SIZE:
            windowSize = self.MAX_WINDOW_SIZE
        self.traceScene.setSceneRect(self.traceScene.itemsBoundingRect()) #resize
        self.setFixedWidth(windowSize)

        # self.listScene.setSceneRect(self.listScene.itemsBoundingRect()) #resize
        self.multiView.setFixedWidth(windowSize)
        cellWidth = windowSize // 2
        self.listView.setColumnWidth(0, cellWidth)
        self.listView.setColumnWidth(1, cellWidth)
        self.listView.setFixedHeight(self.multiView.height() // 4)
        self.multiTraceList.setColumnWidth(0, cellWidth)
        self.multiTraceList.setColumnWidth(1, cellWidth)
        self.view.setFixedWidth(windowSize)

        self.show()

    def _populate_trace_table(self, view, trace_ids):
        numIDs = len(trace_ids)
        view.clearContents()
        view.setRowCount(numIDs)
        row = 0 #start after label row
        for traceID in trace_ids:
            inputID = self.multi_trace.am_obj.get_input_id_for_trace_id(traceID)
            if inputID is None:
                self.workspace.log("No inputID found for trace %s" % traceID)
            view.setItem(row, 0, QTableWidgetItem(traceID))
            view.setItem(row, 1, QTableWidgetItem(inputID))
            row += 1

    def _refresh_heatmap(self):
        multiTrace = self.multi_trace.am_obj
        multiTrace.clear_heatmap()
        multiTrace.is_active_tab = True

        selected_items = self.multiTraceList.selectedItems()
        self._selected_traces.clear()
        for row in range(self.multiTraceList.rowCount()):
            item = self.multiTraceList.item(row, 0)
            if item in selected_items:
                self._selected_traces.append(item.text())
        multiTrace.reload_heatmap(self._selected_traces)
        self.multi_trace.am_event()

    def _refresh_multi_list(self):
        multiTrace = self.multi_trace.am_obj
        trace_ids = multiTrace.get_all_trace_ids()

        self.multiTraceList.clearContents()
        self._populate_trace_table(self.multiTraceList, trace_ids)
        if self._selected_traces and self.multiTraceList.rowCount() > 0:
            self.multiTraceList.item(0,0).setSelected(True)
            self.multiTraceList.item(0,1).setSelected(True)
        else:
            for row in range(self.multiTraceList.rowCount()):
                item = self.multiTraceList.item(row, 0)
                inputItem = self.multiTraceList.item(row, 1)
                if item.text() in self._selected_traces:
                    item.setSelected(True)
                    inputItem.setSelected(True)
        self.multi_trace.am_event()


    def _on_tab_change(self):
        # self._reset()
        multiTrace = self.multi_trace.am_obj
        if self.view.currentIndex() == self.MULTI_TRACE:
            multiTrace.is_active_tab = True
            self._refresh_multi_list()
        elif self.view.currentIndex() == self.SINGLE_TRACE:
            multiTrace = self.multi_trace.am_obj
            multiTrace.is_active_tab = False
            self._show_trace_ids()

    def _on_select_ins(self, **kwargs): # pylint: disable=unused-argument
        if self.trace.am_none:
            return

        if self.mark is not None:
            for i in self.mark.childItems():
                self.mark.removeFromGroup(i)
                self.traceScene.removeItem(i)
            self.traceScene.removeItem(self.mark)

        self.mark = QGraphicsItemGroup()
        self.traceScene.addItem(self.mark)

        if self.selected_ins:
            addr = next(iter(self.selected_ins))
            positions = self.trace.get_positions(addr)
            if positions: #if addr is in list of positions
                if not self._use_precise_position: #handle case where insn was selected from disas view
                    self.curr_position = positions[0] - self.trace.count
                for p in positions:
                    color = self._get_mark_color(p, self.trace.count)
                    y = self._get_mark_y(p)

                    if p == self.trace.count + self.curr_position: #add thicker line for 'current' mark
                        self.mark.addToGroup(self.traceScene.addRect(self.MARK_X, y, self.MARK_WIDTH,
                                            self.MARK_HEIGHT*4, QPen(QColor('black')), QBrush(color)))
                    else:
                        self.mark.addToGroup(self.traceScene.addRect(self.MARK_X, y, self.MARK_WIDTH,
                                                                self.MARK_HEIGHT, QPen(color), QBrush(color)))

                self.traceScene.update() #force redraw of the traceScene
                self.scroll_to_position(self.curr_position)

    def scroll_to_position(self, position):
        relative_pos = self.trace.count + position
        y_offset = self._get_mark_y(relative_pos)

        scrollValue = 0
        if y_offset > 0.5 * self.traceView.size().height():
            scrollValue = y_offset - 0.5 * self.traceView.size().height()
        scrollValue = min(scrollValue, self.traceView.verticalScrollBar().maximum())
        self.traceView.verticalScrollBar().setValue(scrollValue)
        self._use_precise_position = False

    def jump_next_insn(self):
        if self.curr_position + self.trace.count < self.trace.count - 1: #for some reason indexing is done backwards
            self.curr_position += 1
            self._use_precise_position = True
            bbl_addr = self.trace.get_bbl_from_position(self.curr_position)
            func = self.trace.get_func_from_position(self.curr_position)
            self._jump_bbl(func, bbl_addr)

    def jump_prev_insn(self):
        if self.curr_position + self.trace.count > 0:
            self.curr_position -= 1
            self._use_precise_position = True
            bbl_addr = self.trace.get_bbl_from_position(self.curr_position)
            func = self.trace.get_func_from_position(self.curr_position)
            self._jump_bbl(func, bbl_addr)

    def eventFilter(self, obj, event): #specifically to catch arrow keys #pylint: disable=unused-argument
        # more elegant solution to link w/ self.view's scroll bar keypressevent?
        if event.type() == QEvent.Type.KeyPress:
            if not event.modifiers() & Qt.ShiftModifier: #shift + arrowkeys
                return False
            key = event.key()
            if key in [Qt.Key_Up, Qt.Key_Left]:
                self.jump_prev_insn()
            elif key in [Qt.Key_Down, Qt.Key_Right]:
                self.jump_next_insn()
            return True

        return False  # pass through all other events

    def mousePressEvent(self, event):
        button = event.button()
        pos = self._to_logical_pos(event.pos())
        if button == Qt.LeftButton and self.view.currentIndex() == self.SINGLE_TRACE and self._at_legend(pos):
            func = self._get_func_from_y(pos.y())
            bbl_addr = self._get_bbl_from_y(pos.y())
            self._use_precise_position = True
            self.curr_position = self._get_position(pos.y())
            self._jump_bbl(func, bbl_addr)

    def _jump_bbl(self, func, bbl_addr):
        all_insn_addrs = self.workspace.main_instance.project.factory.block(bbl_addr).instruction_addrs
        # TODO: replace this with am_events perhaps?
        if func is None:
            return
        self.workspace.on_function_selected(func)
        self.selected_ins.clear()
        self.selected_ins.update(all_insn_addrs)
        self.selected_ins.am_event()
        # TODO: this ought to happen automatically as a result of the am_event
        self.disasm_view.current_graph.show_instruction(bbl_addr)

    def _get_mark_color(self, i, total):
        relative_gradient_pos = i * 1000 // total
        return self.legend_img.pixelColor(self.LEGEND_WIDTH // 2,
                                          relative_gradient_pos)

    def _get_mark_y(self, i):
        return self.TRACE_FUNC_Y + self.trace_func_unit_height * i


    def _show_trace_ids(self):
        trace_ids = self.multi_trace.get_all_trace_ids()
        # traceID = self.listScene.addText(id_txt, QFont("Source Code Pro", 7))
        # traceID.setPos(5,5)
        self.listView.clearContents()
        self._populate_trace_table(self.listView, trace_ids)
        if len(self.listView.selectedItems()) <= 0 and not self.trace.am_none:
            for row in range(self.listView.rowCount()):
                item = self.listView.item(row, 0)
                inputItem = self.listView.item(row, 1)
                if self.trace.id in item.text():
                    item.setSelected(True)
                    inputItem.setSelected(True)
                    break



    def _show_trace_func(self, show_func_tag):
        x = self.TRACE_FUNC_X
        y = self.TRACE_FUNC_Y
        prev_name = None
        for position in self.trace.trace_func:
            bbl_addr = position.bbl_addr
            func_name = position.func_name
            l.debug('Draw function %x, %s', bbl_addr, func_name)
            color = self.trace.get_func_color(func_name)
            self.trace_func.addToGroup(self.traceScene.addRect(x, y,
                                                          self.TRACE_FUNC_WIDTH, self.trace_func_unit_height,
                                                          QPen(color), QBrush(color)))
            if show_func_tag is True and func_name != prev_name:
                tag = self.traceScene.addText(func_name, QFont("Source Code Pro", 7))
                tag.setPos(x + self.TRACE_FUNC_WIDTH +
                           self.TAG_SPACING, y -
                           tag.boundingRect().height() // 2)
                self.trace_func.addToGroup(tag)
                anchor = self.traceScene.addLine(
                    self.TRACE_FUNC_X + self.TRACE_FUNC_WIDTH, y,
                    x + self.TRACE_FUNC_WIDTH + self.TAG_SPACING, y)
                self.trace_func.addToGroup(anchor)
                prev_name = func_name
            y += self.trace_func_unit_height

    @staticmethod
    def _make_legend_gradient(x1, y1, x2, y2):
        gradient = QLinearGradient(x1, y1, x2, y2)
        gradient.setColorAt(0.0, Qt.red)
        gradient.setColorAt(0.4, Qt.yellow)
        gradient.setColorAt(0.6, Qt.green)
        gradient.setColorAt(0.8, Qt.blue)
        gradient.setColorAt(1.0, Qt.darkBlue)
        return gradient


    def _show_legend(self):
        pen = QPen(Qt.transparent)

        gradient = self._make_legend_gradient(self.LEGEND_X, self.LEGEND_Y,
                                   self.LEGEND_X, self.LEGEND_Y + self.legend_height)
        brush = QBrush(gradient)
        self.legend = self.traceScene.addRect(self.LEGEND_X, self.LEGEND_Y,
                                         self.LEGEND_WIDTH, self.legend_height, pen, brush)

        reference_gradient = self._make_legend_gradient(0, 0, self.LEGEND_WIDTH, 1000)
        base_img = QImage(self.LEGEND_WIDTH, 1000, QImage.Format.Format_ARGB32)
        p = QPainter(base_img)
        p.fillRect(base_img.rect(),reference_gradient)
        self.legend_img = base_img #reference shade


    def _set_mark_color(self):
        for p in range(self.trace.count):
            color = self._get_mark_color(p, self.trace.count)
            self.trace.set_mark_color(p, color)

    def _at_legend(self, pos):
        x = pos.x()
        y = pos.y()
        return self.TRACE_FUNC_X + self.LEGEND_X < x < self.traceView.width() and \
           self.TRACE_FUNC_Y < y < self.TRACE_FUNC_Y + self.legend_height

    def _to_logical_pos(self, pos):
        x_offset = self.traceView.horizontalScrollBar().value()
        y_offset = self.traceView.verticalScrollBar().value()
        return QPoint(pos.x() + x_offset, pos.y() + y_offset)

    def _get_position(self, y):
        y_relative = y - self.legend_height - self.TAB_HEADER_SIZE

        return int(y_relative // self.trace_func_unit_height)

    def _get_bbl_from_y(self, y):
        position = self._get_position(y)
        return self.trace.get_bbl_from_position(position)

    def _get_func_from_y(self, y):
        position = self._get_position(y)
        func = self.trace.get_func_from_position(position)
        return func



import bisect
import functools
import logging
import os
from typing import Dict, List, Tuple, Optional

from angr.errors import SimEngineError


l = logging.getLogger(__name__)

class ObjectAndBase:
    """
    Groups an object and its base address.
    """

    __slots__ = ('obj_name', 'base_addr', 'proj_base_addr', )

    def __init__(self, obj_name: str, base_addr: int, proj_base_addr: Optional[int]):
        self.obj_name = obj_name
        self.base_addr = base_addr
        self.proj_base_addr = proj_base_addr

    def __lt__(self, other):
        if isinstance(other, ObjectAndBase):
            return self.base_addr < other.base_addr
        elif isinstance(other, int):
            return self.base_addr < other
        raise TypeError("Unsupported type %s" % type(other))


def _valid_addr(addr, project):
    if not addr and addr != 0:
        return None
    try:
        return project.loader.find_object_containing(addr)
    except SimEngineError:
        return None

@functools.lru_cache(1024)
def _find_object_base_in_project(object_name, project):
    base_addr = None
    base_obj_name = os.path.basename(object_name)
    for obj in project.loader.all_objects:
        if not hasattr(obj, "binary"):
            continue
        if obj.binary and os.path.basename(obj.binary) == base_obj_name:
            # found it!
            # we assume binary names are unique. if they are not, then add the logic here.
            base_addr = obj.mapped_base
            break
    if base_addr is None:
        l.warning("Cannot find object %s in angr project. Maybe it has not been loaded. Exclude it from the trace.",
                  object_name)
    return base_addr


# cache the last index
last_obj_idx: Optional[int] = None


def _find_obj_in_mapping(addr, mapping) -> Tuple[int,Optional[ObjectAndBase]]:
    idx = bisect.bisect_left(mapping, addr)
    obj = None
    if 0 <= idx < len(mapping):
        # check if addr == object.base
        obj = mapping[idx]
        if addr == obj.base_addr:
            # found
            pass
        elif idx > 0:
            idx = idx - 1
            obj = mapping[idx]
        else:  # idx == len(mapping)
            idx = idx - 1
            obj = mapping[idx]
    return idx, obj


def _apply_trace_offset(addr, mapping, project_baddr, runtime_baddr):
    global last_obj_idx

    if mapping is not None and mapping:
        # find the base address that this address belongs to
        if last_obj_idx is None or last_obj_idx >= len(mapping):
            idx, obj = _find_obj_in_mapping(addr, mapping)
            last_obj_idx = idx
        elif addr < mapping[last_obj_idx].base_addr:
            # find again
            idx, obj = _find_obj_in_mapping(addr, mapping)
            last_obj_idx = idx
        elif addr >= mapping[last_obj_idx + 1].base_addr:
            # find again
            idx, obj = _find_obj_in_mapping(addr, mapping)
            last_obj_idx = idx
        else:
            obj = mapping[last_obj_idx]

        if obj is not None:
            project_base_addr = obj.proj_base_addr
            if project_base_addr is not None:
                return addr + (project_base_addr - obj.base_addr)
            else:
                # not found - the object is probably not loaded in angr? ignore it
                return None

    # fall back
    if project_baddr is not None:
        offset = project_baddr - runtime_baddr
        return addr + offset
    else:
        # this object is probably created before an angr project is created. just give up.
        return None


def trace_to_bb_addrs(trace, project, trace_base):
    """
    convert the trace object to a list of basic blocks, using the given angr project
    """
    bbl_addrs = trace["bb_addrs"]

    mapping: Optional[List[ObjectAndBase]] = None
    if 'map' in trace:
        map_dict: Dict[str,int] = trace["map"]
        mapping = [ObjectAndBase(name, base_addr, _find_object_base_in_project(name, project.am_obj))
                   for name, base_addr in map_dict.items()]
        mapping = list(sorted(mapping, key=lambda o: o.base_addr))  # sort it based on base addresses

    if project.am_none:
        project_baddr = None
    else:
        # only used if self.mapping is not available
        project_baddr = project.loader.main_object.mapped_base
    runtime_baddr = trace_base # this will not be used if self.mapping is available

    # convert over all the trace adders using info from the trace
    to_return = filter(lambda a: _valid_addr(a, project), [
        _apply_trace_offset(addr, mapping, project_baddr, runtime_baddr) for addr in bbl_addrs
    ])
    return list(to_return)




from .job import Job


class SimgrExploreJob(Job):
    def __init__(self, simgr, find=None, avoid=None, step_callback=None, until_callback=None, callback=None):
        super().__init__('Simulation manager exploring')
        self._simgr = simgr
        self._find = find
        self._avoid = avoid
        self._callback = callback
        self._step_callback = step_callback
        self._until_callback = until_callback
        self._interrupted = False

    def _run(self, inst):
        """Run the job. Runs in the worker thread."""
        def until_callback(*args, **kwargs):
            return self._interrupted or callable(self._until_callback) and self._until_callback(*args, **kwargs)
        self._simgr.explore(find=self._find, avoid=self._avoid, step_func=self._step_callback, until=until_callback)
        return self._simgr

    def finish(self, inst, result):
        super().finish(inst, result)
        self._callback(result)

    def __repr__(self):
        return "Exploring %r" % self._simgr

    def keyboard_interrupt(self):
        """Called from GUI thread. Worker thread will check self._interrupted periodically and exit the job early if
        needed. """
        self._interrupted = True

    @classmethod
    def create(cls, simgr, **kwargs):
        def callback(result):
            simgr.am_event(src='job_done', job='explore', result=result)
        return cls(simgr, callback=callback, **kwargs)



from .view import BaseView
from .disassembly_view import DisassemblyView
from .symexec_view import SymexecView
from .functions_view import FunctionsView
from .states_view import StatesView
from .strings_view import StringsView
from .console_view import ConsoleView
from .code_view import CodeView
from .interaction_view import InteractionView
from .patches_view import PatchesView
from .dep_view import DependencyView
from .proximity_view import ProximityView
from .types_view import TypesView
from .hex_view import HexView
from .log_view import LogView
from .data_dep_view import DataDepView
from .registers_view import RegistersView
from .stack_view import StackView
from .traces_view import TracesView
from .trace_map_view import TraceMapView
from .breakpoints_view import BreakpointsView
from .call_explorer_view import CallExplorerView



from typing import List, Iterator
from angr.sim_manager import SimulationManager
from angrmanagement.ui.widgets.qinst_annotation import QInstructionAnnotation, QPassthroughCount
from angrmanagement.ui.widgets.qblock import QBlock
from angrmanagement.plugins import BasePlugin


class SamplePlugin(BasePlugin):
    def __init__(self, workspace):
        super().__init__(workspace)

        workspace.main_instance.register_container('bookmarks', lambda: [], List[int], 'Bookmarked addresses')

    MENU_BUTTONS = ('Add Bookmark',)

    def build_context_menu_functions(self, funcs): # pylint: disable=unused-argument
        yield ("owo", [("uwu", lambda: None), ("o_O", lambda: None)])

    def step_callback(self, simgr: SimulationManager):
        print("Active States: %s" % simgr)

    def build_qblock_annotations(self, qblock: QBlock) -> Iterator[QInstructionAnnotation]:
        return [
            QPassthroughCount(qblock.addr,"entry")
        ]



from PySide6.QtWidgets import QDockWidget


class QSmartDockWidget(QDockWidget):
    """
    Wrapper class for dock widgets.
    """

    def __init__(self, caption, parent=None, on_close=None, on_raise=None):
        super().__init__(caption, parent)

        self.old_size = None
        self.original_min = None
        self.original_max = None
        self._on_close = on_close
        self._on_raise = on_raise

    def restore_original_size(self):

        if self.original_min is None or self.original_max is None:
            return

        self.setMinimumWidth(self.original_min.width())
        self.setMinimumHeight(self.original_min.height())
        self.setMaximumWidth(self.original_max.width())
        self.setMaximumHeight(self.original_max.height())

    def closeEvent(self, event):  # pylint: disable=unused-argument
        if self._on_close is not None:
            self._on_close()
        self.widget().close()

    def raise_(self):
        if self._on_raise is not None:
            self._on_raise()
        super().raise_()



from PySide6.QtGui import QColor


COLOR_SCHEMES = {
    'Light': {
        'disasm_view_operand_color': QColor(0x00, 0x00, 0x80),
        'disasm_view_operand_constant_color': QColor(0x00, 0x00, 0x80),
        'disasm_view_variable_label_color': QColor(0x00, 0x80, 0x00),
        'disasm_view_operand_highlight_color': QColor(0xfc, 0xef, 0x00),
        'disasm_view_operand_select_color': QColor(0xff, 0xff, 0x00),
        'disasm_view_function_color': QColor(0x00, 0x00, 0xff),
        'disasm_view_ir_default_color': QColor(0x80, 0x80, 0x80),
        'disasm_view_label_color': QColor(0x00, 0x00, 0xff),
        'disasm_view_label_highlight_color': QColor(0xf0, 0xf0, 0xbf),
        'disasm_view_target_addr_color': QColor(0x00, 0x00, 0xff),
        'disasm_view_antitarget_addr_color': QColor(0xff, 0x00, 0x00),
        'disasm_view_node_shadow_color': QColor(0x00, 0x00, 0x00, 0x00),
        'disasm_view_node_background_color': QColor(0xfa, 0xfa, 0xfa),
        'disasm_view_node_zoomed_out_background_color': QColor(0xda, 0xda, 0xda),
        'disasm_view_node_border_color': QColor(0xf0, 0xf0, 0xf0),
        'disasm_view_node_instruction_selected_background_color': QColor(0xb8, 0xc3, 0xd6),
        'disasm_view_node_address_color': QColor(0x00, 0x00, 0x00),
        'disasm_view_node_mnemonic_color': QColor(0x00, 0x00, 0x80),
        'disasm_view_selected_node_border_color': QColor(0x6b, 0x71, 0x7c),
        'disasm_view_printable_byte_color': QColor(0x00, 0x80, 0x40),
        'disasm_view_printable_character_color': QColor(0x00, 0x80, 0x40),
        'disasm_view_unprintable_byte_color': QColor(0x80, 0x40, 0x00),
        'disasm_view_unprintable_character_color': QColor(0x80, 0x40, 0x00),
        'disasm_view_unknown_byte_color': QColor(0xf0, 0x00, 0x00),
        'disasm_view_unknown_character_color': QColor(0xf0, 0x00, 0x00),
        'disasm_view_back_edge_color': QColor(0xf9, 0xd5, 0x77),
        'disasm_view_true_edge_color': QColor(0x79, 0xcc, 0xcd),
        'disasm_view_false_edge_color': QColor(0xf1, 0x66, 0x64),
        'disasm_view_direct_jump_edge_color': QColor(0x56, 0x5a, 0x5c),
        'disasm_view_exception_edge_color': QColor(0xf9, 0x91, 0x0a),
        'hex_view_selection_color': QColor(0xFF, 0x00, 0x00),
        'hex_view_selection_alt_color': QColor(0xA0, 0xA0, 0xA4),
        'hex_view_data_color': QColor(0x00, 0x00, 0xFF),
        'hex_view_string_color': QColor(0x00, 0xFF, 0xFF),
        'hex_view_instruction_color': QColor(0xFF, 0x00, 0xFF),
        'function_table_color': QColor(0x00, 0x00, 0x00),
        'function_table_syscall_color': QColor(0x00, 0x00, 0x80),
        'function_table_plt_color': QColor(0x00, 0x80, 0x00),
        'function_table_simprocedure_color': QColor(0x80, 0x00, 0x00),
        'function_table_alignment_color': QColor(0x80, 0x00, 0x80),
        'function_table_signature_bg_color': QColor(0xaa, 0xff, 0xff),
        'palette_window': QColor(0xef, 0xef, 0xef, 0xff),
        'palette_windowtext': QColor(0x00, 0x00, 0x00, 0xff),
        'palette_base': QColor(0xff, 0xff, 0xff, 0xff),
        'palette_alternatebase': QColor(0xf7, 0xf7, 0xf7, 0xff),
        'palette_tooltipbase': QColor(0xff, 0xff, 0xdc, 0xff),
        'palette_tooltiptext': QColor(0x00, 0x00, 0x00, 0xff),
        'palette_text': QColor(0x00, 0x00, 0x00, 0xff),
        'palette_button': QColor(0xef, 0xef, 0xef, 0xff),
        'palette_buttontext': QColor(0x00, 0x00, 0x00, 0xff),
        'palette_brighttext': QColor(0xff, 0xff, 0xff, 0xff),
        'palette_highlight': QColor(0x30, 0x8c, 0xc6, 0xff),
        'palette_highlightedtext': QColor(0xff, 0xff, 0xff, 0xff),
        'palette_disabled_text': QColor(0xbe, 0xbe, 0xbe, 0xff),
        'palette_disabled_buttontext': QColor(0xbe, 0xbe, 0xbe, 0xff),
        'palette_disabled_windowtext': QColor(0xbe, 0xbe, 0xbe, 0xff),
        'palette_light': QColor(0xff, 0xff, 0xff, 0xff),
        'palette_midlight': QColor(0xca, 0xca, 0xca, 0xff),
        'palette_dark': QColor(0x9f, 0x9f, 0x9f, 0xff),
        'palette_mid': QColor(0xb8, 0xb8, 0xb8, 0xff),
        'palette_shadow': QColor(0x76, 0x76, 0x76, 0xff),
        'palette_link': QColor(0x00, 0x00, 0xff, 0xff),
        'palette_linkvisited': QColor(0xff, 0x00, 0xff, 0xff),
        'feature_map_color_regular_function': QColor(0x00, 0xa0, 0xe8),
        'feature_map_color_unknown': QColor(0x0a, 0x0a, 0x0a),
        'feature_map_color_delimiter': QColor(0x00, 0x00, 0x00),
        'feature_map_color_data': QColor(0xc0, 0xc0, 0xc0),
        'pseudocode_comment_color': QColor(0x00, 0x80, 0x00, 0xff),
        'pseudocode_function_color': QColor(0x00, 0x00, 0xff, 0xff),
        'pseudocode_quotation_color': QColor(0x00, 0x80, 0x00, 0xff),
        'pseudocode_keyword_color': QColor(0x00, 0x00, 0x80, 0xff),
        'proximity_node_background_color': QColor(0xfa, 0xfa, 0xfa),
        'proximity_node_selected_background_color': QColor(0xcc, 0xcc, 0xcc),
        'proximity_node_border_color': QColor(0xf0, 0xf0, 0xf0),
        'proximity_function_node_text_color': QColor(0xff, 0x00, 0x00),
        'proximity_string_node_text_color': QColor(0x00, 0x80, 0x00),
        'proximity_integer_node_text_color': QColor(0x00, 0x00, 0x80),
        'proximity_variable_node_text_color': QColor(0x00, 0x00, 0x80),
        'proximity_unknown_node_text_color': QColor(0x00, 0x00, 0x80),
        'proximity_call_node_text_color': QColor(0x00, 0x00, 0xff),
        'proximity_call_node_text_color_plt': QColor(0x8b, 0x00, 0x8b),
        'proximity_call_node_text_color_simproc': QColor(0x8b, 0x00, 0x8b),
    },

    'Dark': {
        'disasm_view_operand_color': QColor(0xf0, 0xf0, 0x5a),
        'disasm_view_operand_constant_color': QColor(0x34, 0xf0, 0x8c),
        'disasm_view_variable_label_color': QColor(0x34, 0xd4, 0xf0),
        'disasm_view_operand_highlight_color': QColor(0x05, 0x2f, 0x50),
        'disasm_view_operand_select_color': QColor(0x09, 0x50, 0x8d),
        'disasm_view_function_color': QColor(0xc8, 0xc8, 0xc8),
        'disasm_view_ir_default_color': QColor(0x80, 0x80, 0x80),
        'disasm_view_label_color': QColor(0x00, 0xaa, 0xff),
        'disasm_view_label_highlight_color': QColor(0x2f, 0x2f, 0x25),
        'disasm_view_target_addr_color': QColor(0x00, 0xaa, 0xff),
        'disasm_view_antitarget_addr_color': QColor(0xff, 0x00, 0x00),
        'disasm_view_node_shadow_color': QColor(0x00, 0x00, 0x00, 0x4b),
        'disasm_view_node_background_color': QColor(0x3c, 0x3c, 0x3c),
        'disasm_view_node_zoomed_out_background_color': QColor(0x64, 0x64, 0x64),
        'disasm_view_node_border_color': QColor(0x50, 0x50, 0x50),
        'disasm_view_node_instruction_selected_background_color': QColor(0x4c, 0x50, 0x58),
        'disasm_view_node_address_color': QColor(0x2c, 0xc9, 0x76),
        'disasm_view_node_mnemonic_color': QColor(0xe0, 0xe0, 0xe0),
        'disasm_view_selected_node_border_color': QColor(0x6b, 0x71, 0x7c),
        'disasm_view_printable_byte_color': QColor(0x00, 0x80, 0x40),
        'disasm_view_printable_character_color': QColor(0x00, 0x80, 0x40),
        'disasm_view_unprintable_byte_color': QColor(0xba, 0xba, 0xba),
        'disasm_view_unprintable_character_color': QColor(0xba, 0xba, 0xba),
        'disasm_view_unknown_byte_color': QColor(0xf0, 0x00, 0x00),
        'disasm_view_unknown_character_color': QColor(0xf0, 0x00, 0x00),
        'disasm_view_back_edge_color': QColor(0xf9, 0xd5, 0x77),
        'disasm_view_true_edge_color': QColor(0x79, 0xcc, 0xcd),
        'disasm_view_false_edge_color': QColor(0xf1, 0x66, 0x64),
        'disasm_view_direct_jump_edge_color': QColor(0x56, 0x5a, 0x5c),
        'disasm_view_exception_edge_color': QColor(0xf9, 0x91, 0x0a),
        'hex_view_selection_color': QColor(0xFF, 0x00, 0x00),
        'hex_view_selection_alt_color': QColor(0xA0, 0xA0, 0xA4),
        'hex_view_data_color': QColor(0x00, 0x00, 0xFF),
        'hex_view_string_color': QColor(0x00, 0xFF, 0xFF),
        'hex_view_instruction_color': QColor(0xFF, 0x00, 0xFF),
        'function_table_color': QColor(0xe0, 0xe0, 0xe0),
        'function_table_syscall_color': QColor(0x00, 0x00, 0x80),
        'function_table_plt_color': QColor(0x00, 0x80, 0x00),
        'function_table_simprocedure_color': QColor(0x80, 0x00, 0x00),
        'function_table_alignment_color': QColor(0x80, 0x80, 0x00),
        'function_table_signature_bg_color': QColor(0xaa, 0xff, 0xff),
        'palette_window': QColor(0x35, 0x35, 0x35),
        'palette_windowtext': QColor(0xff, 0xff, 0xff),
        'palette_base': QColor(0x28, 0x28, 0x28),
        'palette_alternatebase': QColor(0x35, 0x35, 0x35),
        'palette_tooltipbase': QColor(0x35, 0x35, 0x35),
        'palette_tooltiptext': QColor(0xff, 0xff, 0xff),
        'palette_text': QColor(0xff, 0xff, 0xff),
        'palette_button': QColor(0x35, 0x35, 0x35),
        'palette_buttontext': QColor(0xff, 0xff, 0xff),
        'palette_brighttext': QColor(0xff, 0x00, 0x00),
        'palette_highlight': QColor(0x28, 0x28, 0x28).lighter(),
        'palette_highlightedtext': QColor(0xff, 0xff, 0xff),
        'palette_disabled_text': QColor(0x80, 0x80, 0x80),
        'palette_disabled_buttontext': QColor(0x80, 0x80, 0x80),
        'palette_disabled_windowtext': QColor(0x80, 0x80, 0x80),
        'palette_light': QColor(0x00, 0x00, 0x00),
        'palette_midlight': QColor(0x00, 0x00, 0x00),
        'palette_dark': QColor(0x00, 0x00, 0x00),
        'palette_mid': QColor(0x46, 0x46, 0x46),
        'palette_shadow': QColor(0x00, 0x00, 0x00),
        'palette_link': QColor(0x2d, 0xc5, 0x2d).lighter(),
        'palette_linkvisited': QColor(0x2d, 0xc5, 0x2d).darker(),
        'feature_map_color_regular_function': QColor(0x00, 0xa0, 0xe8),
        'feature_map_color_unknown': QColor(0x0a, 0x0a, 0x0a),
        'feature_map_color_delimiter': QColor(0x00, 0x00, 0x00),
        'feature_map_color_data': QColor(0xc0, 0xc0, 0xc0),
        'pseudocode_comment_color': QColor(0x00, 0x80, 0x00, 0xff),
        'pseudocode_function_color': QColor(0x60, 0x80, 0xff, 0xff),
        'pseudocode_quotation_color': QColor(0x00, 0x80, 0x00, 0xff),
        'pseudocode_keyword_color': QColor(0x00, 0xff, 0xff, 0xff),
        'proximity_node_background_color': QColor(0x3c, 0x3c, 0x3c),
        'proximity_node_selected_background_color': QColor(0x4c, 0x50, 0x58),
        'proximity_node_border_color': QColor(0x50, 0x50, 0x50),
        'proximity_function_node_text_color': QColor(0xff, 0x00, 0xff),
        'proximity_string_node_text_color': QColor(0xfa, 0xfa, 0xfa),
        'proximity_integer_node_text_color': QColor(0xfa, 0xfa, 0xfa),
        'proximity_variable_node_text_color': QColor(0xfa, 0xfa, 0xfa),
        'proximity_unknown_node_text_color': QColor(0xfa, 0xfa, 0xfa),
        'proximity_call_node_text_color': QColor(0x34, 0xd4, 0xf0),
        'proximity_call_node_text_color_plt': QColor(0x2c, 0xc9, 0x76),
        'proximity_call_node_text_color_simproc': QColor(0x2c, 0xc9, 0x76),
    }
}



# pylint:disable=no-self-use
import os
import logging
import pickle
import sys
import time
from functools import partial
from typing import Optional, TYPE_CHECKING

from PySide6.QtWidgets import QMainWindow, QTabWidget, QFileDialog, QProgressBar, QProgressDialog
from PySide6.QtWidgets import QMessageBox, QTabBar
from PySide6.QtGui import QIcon, QDesktopServices, QKeySequence, QShortcut
from PySide6.QtCore import Qt, QSize, QEvent, QUrl

import angr
import angr.flirt

try:
    from angr.angrdb import AngrDB
except ImportError:
    AngrDB = None  # type: Optional[type]

try:
    import archr
    import keystone
except ImportError:
    archr = None
    keystone = None

from ..daemon import daemon_exists, run_daemon_process, daemon_conn
from ..daemon.client import ClientService
from ..logic import GlobalInfo
from ..data.instance import Instance
from ..data.library_docs import LibraryDocs
from ..data.jobs.loading import LoadTargetJob, LoadBinaryJob, LoadAngrDBJob
from ..data.jobs import DependencyAnalysisJob
from ..config import IMG_LOCATION, Conf, save_config
from ..utils.io import isurl, download_url
from ..utils.env import is_pyinstaller, app_root
from ..errors import InvalidURLError, UnexpectedStatusCodeError
from .menus.file_menu import FileMenu
from .menus.analyze_menu import AnalyzeMenu
from .menus.help_menu import HelpMenu
from .menus.view_menu import ViewMenu
from .menus.plugin_menu import PluginMenu
from .workspace import Workspace
from .dialogs.load_plugins import LoadPlugins
from .dialogs.load_docker_prompt import LoadDockerPrompt, LoadDockerPromptError
from .dialogs.new_state import NewState
from .dialogs.about import LoadAboutDialog
from .dialogs.preferences import Preferences
from .toolbars import FileToolbar, DebugToolbar
from .toolbar_manager import ToolbarManager

if TYPE_CHECKING:
    from PySide6.QtWidgets import QApplication

_l = logging.getLogger(name=__name__)


class MainWindow(QMainWindow):
    """
    The main window of angr management.
    """

    def __init__(self, app: Optional['QApplication'] = None, parent=None, show=True, use_daemon=False):
        super().__init__(parent)

        icon_location = os.path.join(IMG_LOCATION, 'angr.png')
        self.setWindowIcon(QIcon(icon_location))

        GlobalInfo.main_window = self
        self.shown_at_start = show

        # initialization
        self.setMinimumSize(QSize(400, 400))
        self.setDockNestingEnabled(True)

        self.app: Optional['QApplication'] = app
        self.workspace: Workspace = None
        self.central_widget: QMainWindow = None

        self.toolbar_manager: ToolbarManager = ToolbarManager(self)
        self._progressbar = None  # type: QProgressBar
        self._progress_dialog = None # type: QProgressDialog

        self.defaultWindowFlags = None

        # menus
        self._file_menu = None  # FileMenu
        self._analyze_menu = None
        self._view_menu = None
        self._help_menu = None
        self._plugin_menu = None

        self._init_statusbar()
        self._init_workspace()
        self._init_toolbars()
        self._init_menus()
        self._init_plugins()
        self._init_library_docs()
        self._init_url_scheme_handler()

        self.workspace.plugins.on_workspace_initialized(self)

        self._init_shortcuts()
        self._init_flirt_signatures()

        self._run_daemon(use_daemon=use_daemon)

        # I'm ready to show off!
        if show:
            self.showMaximized()
            self.windowHandle().screenChanged.connect(self.on_screen_changed)
            self.show()

        self.status = "Ready."

    def sizeHint(self, *args, **kwargs):  # pylint: disable=unused-argument,no-self-use
        return QSize(1200, 800)

    #
    # Properties
    #

    @property
    def caption(self):
        return self.getWindowTitle()

    @caption.setter
    def caption(self, v):
        self.setWindowTitle(v)

    #
    # Dialogs
    #

    def open_mainfile_dialog(self):
        # pylint: disable=assigning-non-slot
        # https://github.com/PyCQA/pylint/issues/3793
        file_path, _ = QFileDialog.getOpenFileName(self, "Open a binary", Conf.last_used_directory,
                                                   "All executables (*);;"
                                                   "Windows PE files (*.exe);;"
                                                   "Core Dumps (*.core);;"
                                                   "angr database (*.adb)",
                                                   )
        Conf.last_used_directory = os.path.dirname(file_path)
        return file_path

    def _pick_image_dialog(self):
        try:
            prompt = LoadDockerPrompt(parent=self)
        except LoadDockerPromptError:
            return None
        if prompt.exec_() == 0:
            return None  # User canceled
        return prompt.textValue()

    def open_load_plugins_dialog(self):
        dlg = LoadPlugins(self.workspace.plugins)
        dlg.setModal(True)
        dlg.exec_()

    def open_newstate_dialog(self):
        if self.workspace.main_instance.project.am_none:
            QMessageBox.critical(self,
                                 "Cannot create new states",
                                 "Please open a binary to analyze first.")
            return
        new_state_dialog = NewState(self.workspace.main_instance, parent=self, create_simgr=True)
        new_state_dialog.exec_()

    def open_doc_link(self):
        QDesktopServices.openUrl(QUrl("https://docs.angr.io/", QUrl.TolerantMode))

    def open_about_dialog(self):
        dlg = LoadAboutDialog()
        dlg.exec_()

    #
    # Widgets
    #

    def _init_statusbar(self):
        self._progressbar = QProgressBar()

        self._progressbar.setMinimum(0)
        self._progressbar.setMaximum(100)
        self._progressbar.hide()

        self.statusBar().addPermanentWidget(self._progressbar)

        self._progress_dialog = QProgressDialog("Waiting...", "Cancel", 0, 100, self)
        self._progress_dialog.setAutoClose(False)
        self._progress_dialog.setWindowFlags(self._progress_dialog.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self._progress_dialog.setModal(True)
        self._progress_dialog.setMinimumDuration(2**31 - 1)
        def on_cancel():
            if self.workspace is None:
                return
            for job in self.workspace.main_instance.jobs:
                if job.blocking:
                    job.keyboard_interrupt()
                    break
        self._progress_dialog.canceled.connect(on_cancel)
        self._progress_dialog.close()

    def _init_toolbars(self):
        for cls in (FileToolbar, DebugToolbar):
            self.toolbar_manager.show_toolbar_by_class(cls)

    #
    # Menus
    #

    def _init_menus(self):
        self._file_menu = FileMenu(self)
        self._analyze_menu = AnalyzeMenu(self)
        self._view_menu = ViewMenu(self)
        self._help_menu = HelpMenu(self)
        self._plugin_menu = PluginMenu(self)

        for path in Conf.recent_files:
            self._file_menu.add_recent(path)

        # TODO: Eventually fix menu bars to have native support on MacOS
        # if on a Mac, don't use the native menu bar (bug mitigation from QT)
        if sys.platform == 'darwin':
            self.menuBar().setNativeMenuBar(False)

        self.menuBar().addMenu(self._file_menu.qmenu())
        self.menuBar().addMenu(self._view_menu.qmenu())
        self.menuBar().addMenu(self._analyze_menu.qmenu())
        self.menuBar().addMenu(self._plugin_menu.qmenu())
        self.menuBar().addMenu(self._help_menu.qmenu())

    #
    # Workspace
    #

    def _init_workspace(self):
        """
        Initialize workspace

        :return:    None
        """
        self.central_widget = QMainWindow()
        self.setCentralWidget(self.central_widget)
        wk = Workspace(self, Instance())
        self.workspace = wk
        self.workspace.view_manager.tabify_center_views()
        self.central_widget.setTabPosition(Qt.RightDockWidgetArea, QTabWidget.North)
        self.central_widget.setDockNestingEnabled(True)

        def set_caption(**kwargs):  # pylint: disable=unused-argument
            if self.workspace.main_instance.project.am_none:
                self.caption = ''
            elif self.workspace.main_instance.project.filename is None:
                self.caption = "Loaded from stream"
            else:
                self.caption = os.path.basename(self.workspace.main_instance.project.filename)

        self.workspace.main_instance.project.am_subscribe(set_caption)

        self.tab = self.central_widget.findChild(QTabBar)
        self.tab.tabBarClicked.connect(self.on_center_tab_clicked)

    #
    # Shortcuts
    #

    def interrupt_current_job(self):
        self.workspace.main_instance.interrupt_current_job()

    def _init_shortcuts(self):
        """
        Initialize shortcuts

        :return:    None
        """

        center_dockable_views = self.workspace.view_manager.get_center_views()
        for i in range(1, len(center_dockable_views) + 1):
            QShortcut(QKeySequence('Ctrl+' + str(i)), self, center_dockable_views[i - 1].raise_)

        QShortcut(QKeySequence("Ctrl+I"), self, self.interrupt_current_job)

        # Raise the DisassemblyView after everything has initialized
        center_dockable_views[0].raise_()

        # Toggle exec breakpoint
        QShortcut(QKeySequence(Qt.Key_F2), self, self.workspace.toggle_exec_breakpoint)

        # Single step
        QShortcut(QKeySequence(Qt.Key_F7), self, self.workspace.step_forward)

        # Run
        QShortcut(QKeySequence(Qt.Key_F9), self, self.workspace.continue_forward)

    #
    # Plugins
    #

    def _init_plugins(self):
        self.workspace.plugins.discover_and_initialize_plugins()

    #
    # FLIRT Signatures
    #

    def _init_flirt_signatures(self):
        if Conf.flirt_signatures_root:
            # if it's a relative path, it's relative to the angr-management package
            if os.path.isabs(Conf.flirt_signatures_root):
                flirt_signatures_root = Conf.flirt_signatures_root
            else:
                if is_pyinstaller():
                    flirt_signatures_root = os.path.join(app_root(), Conf.flirt_signatures_root)
                else:
                    # when running as a Python package, we should use the git submodule, which is on the same level
                    # with (instead of inside) the angrmanagement module directory.
                    flirt_signatures_root = os.path.join(app_root(), "..", Conf.flirt_signatures_root)
            flirt_signatures_root = os.path.normpath(flirt_signatures_root)
            _l.info("Loading FLIRT signatures from %s.", flirt_signatures_root)
            angr.flirt.load_signatures(flirt_signatures_root)

    #
    # Library docs
    #

    def _init_library_docs(self):
        GlobalInfo.library_docs = LibraryDocs()
        if Conf.library_docs_root:
            GlobalInfo.library_docs.load_func_docs(Conf.library_docs_root)

    #
    # Daemon
    #

    def _run_daemon(self, use_daemon=None):

        if use_daemon is None:
            # Load it from the configuration file
            use_daemon = Conf.use_daemon

        if not use_daemon:
            return

        # connect to daemon (if there is one)
        if not daemon_exists():
            print("[+] Starting a new daemon.")
            run_daemon_process()
            time.sleep(0.2)
        else:
            print("[+] Connecting to an existing angr management daemon.")

        while True:
            try:
                GlobalInfo.daemon_conn = daemon_conn(service=ClientService)
            except ConnectionRefusedError:
                print("[-] Connection failed... try again.")
                time.sleep(0.4)
                continue
            print("[+] Connected to daemon.")
            break

        from rpyc import BgServingThread  # pylint:disable=import-outside-toplevel
        _ = BgServingThread(GlobalInfo.daemon_conn)

    #
    # URL scheme handler setup
    #

    def _init_url_scheme_handler(self):
        # URL scheme
        from ..logic.url_scheme import AngrUrlScheme  # pylint:disable=import-outside-toplevel

        scheme = AngrUrlScheme()
        registered, _ = scheme.is_url_scheme_registered()
        supported = scheme.is_url_scheme_supported()
        checrs_plugin = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
        if checrs_plugin is None:
            return

        if not registered and supported:
            btn = QMessageBox.question(None, "Setting up angr URL scheme",
                                       "angr URL scheme allows \"deep linking\" from browsers and other applications "
                                       "by registering the angr:// protocol to the current user. Do you want to "
                                       "register it? You may unregister at any "
                                       "time in Preferences.",
                                       defaultButton=QMessageBox.Yes)
            if btn == QMessageBox.Yes:
                try:
                    AngrUrlScheme().register_url_scheme()
                except (ValueError, FileNotFoundError) as ex:
                    QMessageBox.warning(None, "Error in registering angr URL scheme",
                                        "Failed to register the angr URL scheme.\n"
                                        "The following exception occurred:\n"
                                        + str(ex))

    #
    # Event
    #

    def closeEvent(self, event):

        # Ask if the user wants to save things
        if self.workspace.main_instance is not None \
                and not self.workspace.main_instance.project.am_none \
                and self.shown_at_start:
            msgbox = QMessageBox()
            msgbox.setWindowTitle("Save database")
            msgbox.setText("angr management is about to exit. Do you want to save the database?")
            msgbox.setIcon(QMessageBox.Question)
            msgbox.setWindowIcon(self.windowIcon())
            msgbox.setStandardButtons(QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel)
            msgbox.setDefaultButton(QMessageBox.Yes)
            r = msgbox.exec_()

            if r == QMessageBox.Cancel:
                event.ignore()
                return
            elif r == QMessageBox.Yes:
                save_r = self.save_database()
                if not save_r:
                    # failed to save the database
                    event.ignore()
                    return

        for plugin in list(self.workspace.plugins.active_plugins.values()):
            self.workspace.plugins.deactivate_plugin(plugin)
        event.accept()

    def event(self, event):

        if event.type() == QEvent.User:
            # our event callback

            try:
                event.result = event.execute()
            except Exception as ex:  # pylint:disable=broad-except
                event.exception = ex
            event.event.set()

            return True

        return super().event(event)

    def on_screen_changed(self, screen):
        """
        When changing from one screen to another, ask disassembly views to refresh in case the DPI is changed.
        """
        self.workspace.current_screen.am_obj = screen
        self.workspace.current_screen.am_event()

    def on_center_tab_clicked(self, index):
        self.workspace.view_manager.handle_center_tab_click(index)

    #
    # Actions
    #

    def reload(self):
        self.workspace.reload()

    def open_file_button(self):
        file_path = self.open_mainfile_dialog()
        if not file_path:
            return
        self.load_file(file_path)

    def open_trace_file_button(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "Open a trace file", Conf.last_used_directory,
                                                   "All files (*);;"
                                                   "Trace files (*.trace);;",
                                                   )
        Conf.last_used_directory = os.path.dirname(file_path)
        if not file_path:
            return
        self.load_trace_file(file_path)

    def open_docker_button(self):
        required = {
            'archr: git clone https://github.com/angr/archr && cd archr && pip install -e .': archr,
            'keystone: pip install --no-binary keystone-engine keystone-engine': keystone
        }
        is_missing = [key for key, value in required.items() if value is None]
        if len(is_missing) > 0:
            req_msg = 'You need to install the following:\n\n\t' + '\n\t'.join(is_missing)
            req_msg += '\n\nInstall them to enable this functionality.'
            req_msg += '\nRelaunch angr-management after install.'
            QMessageBox(self).critical(None, 'Dependency error', req_msg)
            return

        img_name = self._pick_image_dialog()
        if img_name is None:
            return
        target = archr.targets.DockerImageTarget(img_name, target_path=None)
        self.workspace.main_instance.add_job(LoadTargetJob(target))
        self.workspace.main_instance.img_name = img_name

    def load_trace_file(self, file_path):
        if isurl(file_path):
            QMessageBox.critical(self,
                                 "Unsupported Action",
                                 "Downloading trace files is not yet supported."
                                 "Please specify a path to a file on disk.")
        else:
            # File
            if os.path.isfile(file_path):
                try:
                    with open(file_path, 'rb') as f:
                        loaded_sim_state = pickle.load(f)
                        analysis_params = {
                            'end_state': loaded_sim_state,
                            'start_addr': None,
                            'end_addr': None,
                            'block_addrs': None,
                        }
                        self.workspace.view_data_dependency_graph(analysis_params)
                    self._recent_file(file_path)
                except pickle.PickleError:
                    QMessageBox.critical(self,
                                         "Unable to load trace file",
                                         "Trace file must contain a serialized SimState.")
            else:
                QMessageBox.critical(self,
                                     "File not found",
                                     f"angr management cannot open file {file_path}. "
                                     "Please make sure that the file exists.")

    def load_file(self, file_path):

        if not isurl(file_path):
            # file
            if os.path.isfile(file_path):
                if file_path.endswith(".trace"):
                    self.workspace.load_trace_from_path(file_path)
                    return

                self.workspace.main_instance.binary_path = file_path
                self.workspace.main_instance.original_binary_path = file_path
                if file_path.endswith(".adb"):
                    self._load_database(file_path)
                else:
                    self._recent_file(file_path)
                    self.workspace.main_instance.add_job(LoadBinaryJob(file_path))
            else:
                QMessageBox.critical(self,
                                     "File not found",
                                     f"angr management cannot open file {file_path}. "
                                     "Please make sure that the file exists.")
        else:
            # url
            r = QMessageBox.question(self,
                                     "Downloading a file",
                                     f"Do you want to download a file from {file_path} and open it in angr management?",
                                     defaultButton=QMessageBox.Yes)
            if r == QMessageBox.Yes:
                try:
                    target_path = download_url(file_path, parent=self, to_file=True, file_path=None)
                except InvalidURLError:
                    QMessageBox.critical(self,
                                         "Downloading failed",
                                         "angr management failed to download the file. The URL is invalid.")
                    return
                except UnexpectedStatusCodeError as ex:
                    QMessageBox.critical(self,
                                         "Downloading failed",
                                         "angr management failed to retrieve the header of the file. "
                                         f"The HTTP request returned an unexpected status code {ex.status_code}.")
                    return

                if target_path:
                    # open the file - now it's a local file
                    self.workspace.main_instance.binary_path = target_path
                    self.workspace.main_instance.original_binary_path = file_path
                    self.load_file(target_path)

    def load_database(self):
        # Open File window
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Load angr database", ".",
            "angr databases (*.adb)",
        )

        if not file_path:
            return

        self._load_database(file_path)

    def save_database(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return True

        if self.workspace.main_instance.database_path is None:
            return self.save_database_as()
        else:
            return self._save_database(self.workspace.main_instance.database_path)

    def save_database_as(self):

        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        default_database_path = self.workspace.main_instance.database_path
        if default_database_path is None:
            default_database_path = os.path.normpath(self.workspace.main_instance.project.filename) + ".adb"

        # Open File window
        file_path, _ = QFileDialog.getSaveFileName(
            self, "Save angr database", default_database_path,
            "angr databases (*.adb)",
        )

        if not file_path:
            return False

        if not file_path.endswith(".adb"):
            file_path = file_path + ".adb"

        return self._save_database(file_path)

    def load_trace(self):
        file_path, _ = QFileDialog.getOpenFileName(self, "Load trace", ".", "bintrace (*.trace)")
        if not file_path:
            return
        self.workspace.load_trace_from_path(file_path)

    def preferences(self):
        pref = Preferences(self.workspace, parent=self)
        pref.exec_()

    def quit(self):
        self.close()

    def run_variable_recovery(self):
        self.workspace._get_or_create_disassembly_view().variable_recovery_flavor = 'accurate'

    def run_induction_variable_analysis(self):
        self.workspace._get_or_create_disassembly_view().run_induction_variable_analysis()

    def run_dependency_analysis(self, func_addr: Optional[int] = None, func_arg_idx: Optional[int] = None):
        if self.workspace is None or self.workspace.main_instance is None:
            return
        dep_analysis_job = DependencyAnalysisJob(func_addr=func_addr, func_arg_idx=func_arg_idx)
        self.workspace.main_instance.add_job(dep_analysis_job)

    def run_analysis(self):
        if self.workspace:
            self.workspace.main_instance.run_analysis()

    def decompile_current_function(self):
        if self.workspace is not None:
            self.workspace.decompile_current_function()

    def view_proximity_for_current_function(self):
        if self.workspace is not None:
            self.workspace.view_proximity_for_current_function()

    def interact(self):
        self.workspace.interact_program(self.workspace.main_instance.img_name)

    #
    # Other public methods
    #

    def progress(self, status, progress):
        self.statusBar().showMessage(f'Working... {status}')
        self._progress_dialog.setLabelText(status)
        self._progressbar.show()
        self._progressbar.setValue(progress)
        self._progress_dialog.setValue(progress)

    def progress_done(self):
        self._progressbar.hide()
        self.statusBar().showMessage("Ready.")
        self._progress_dialog.hide()

    def bring_to_front(self):
        self.setWindowState((self.windowState() & ~Qt.WindowMinimized) | Qt.WindowActive)
        self.activateWindow()
        self.raise_()

    #
    # Private methods
    #

    def _recent_file(self, file_path):
        file_path = os.path.abspath(file_path)
        self._file_menu.add_recent(file_path)
        Conf.recent_file(file_path)
        save_config()

    def _load_database(self, file_path: str):

        if AngrDB is None:
            QMessageBox.critical(None, 'Error',
                                 'AngrDB is not enabled. Maybe you do not have SQLAlchemy installed?')
            return


        other_kbs = {}
        extra_info = {}

        job = LoadAngrDBJob(file_path, ["global", "pseudocode_variable_kb"], other_kbs=other_kbs, extra_info=extra_info)
        job._on_finish = partial(self._on_load_database_finished, job)
        self.workspace.main_instance.add_job(job)

    def _on_load_database_finished(self, job: LoadAngrDBJob):
        proj = job.project

        if proj is None:
            return

        self._recent_file(job.file_path)

        cfg = proj.kb.cfgs['CFGFast']
        cfb = proj.analyses.CFB()  # it will load functions from kb

        self.workspace.main_instance.database_path = job.file_path

        self.workspace.main_instance._reset_containers()
        self.workspace.main_instance.project = proj
        self.workspace.main_instance.cfg = cfg
        self.workspace.main_instance.cfb = cfb
        if "pseudocode_variable_kb" in job.other_kbs:
            self.workspace.main_instance.pseudocode_variable_kb = job.other_kbs["pseudocode_variable_kb"]
        else:
            self.workspace.main_instance.initialize_pseudocode_variable_kb()
        self.workspace.main_instance.project.am_event(initialized=True)

        # trigger callbacks
        self.workspace.reload()
        self.workspace.on_cfg_generated()
        self.workspace.plugins.angrdb_load_entries(job.extra_info)

    def _save_database(self, file_path):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        if AngrDB is None:
            QMessageBox.critical(None, 'Error',
                                 'AngrDB is not enabled. Maybe you do not have SQLAlchemy installed?')
            return False

        self.workspace.plugins.handle_project_save(file_path)

        angrdb = AngrDB(project=self.workspace.main_instance.project)
        extra_info = self.workspace.plugins.angrdb_store_entries()
        angrdb.dump(file_path, kbs=[
                self.workspace.main_instance.kb,
                self.workspace.main_instance.pseudocode_variable_kb,
                ],
            extra_info=extra_info,
            )

        self.workspace.main_instance.database_path = file_path
        return True



from typing import Optional, TYPE_CHECKING
from collections import OrderedDict

from PySide6.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QLineEdit, QListWidget, \
    QDialogButtonBox
from PySide6.QtGui import Qt
from angr.sim_type import SimType, TypeRef, SimTypePointer, NamedTypeMixin
from angr.analyses.decompiler.structured_codegen.c import CVariable, CFunction, CConstruct, CFunctionCall, CStructField

if TYPE_CHECKING:
    from angrmanagement.ui.views.code_view import CodeView


class NodeNameBox(QLineEdit):
    """
    QLineEdit that validates node names.
    """

    def __init__(self, textchanged_callback, parent=None):
        super().__init__(parent)

        self.textChanged.connect(textchanged_callback)

    @property
    def name(self):
        text = self.text()
        if self._is_valid_node_name(text):
            return text.strip()
        return None

    @staticmethod
    def _is_valid_node_name(name):
        stripped = name.strip()
        return stripped and ' ' not in stripped


class RenameNode(QDialog):
    """
    Dialog for renaming a node.
    """

    def __init__(self, code_view: Optional['CodeView']=None, node: Optional[CConstruct]=None, parent=None):
        super().__init__(parent)

        # initialization
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self._code_view = code_view
        self._node = node

        self._name_box: NodeNameBox = None
        self._status_label = None
        self._ok_button: QPushButton = None
        self._suggestion_box: QListWidget = None

        self.setWindowTitle('Rename')

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        if not isinstance(self._node, CVariable):
            self._suggestion_box.setEnabled(False)
        else:
            if self._node.unified_variable is not None:
                if not self._node.unified_variable.candidate_names:
                    self._suggestion_box.setEnabled(False)
                else:
                    for candidate in self._node.unified_variable.candidate_names:
                        self._suggestion_box.addItem(candidate)

        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self):

        # name label

        name_label = QLabel(self)
        name_label.setText('New name')

        name_box = NodeNameBox(self._on_name_changed, self)
        if self._node is not None:
            # parse node type, either a Function header or a Variable.
            if isinstance(self._node, CVariable) and self._node.unified_variable and self._node.unified_variable.name:
                name_box.setText(self._node.unified_variable.name)
            elif isinstance(self._node, CVariable) and self._node.variable.region == '':
                name_box.setText(self._node.variable.name)
            elif isinstance(self._node, CFunction) and self._node.name:
                name_box.setText(self._node.name)
            elif isinstance(self._node, CFunctionCall):
                name_box.setText(self._node.callee_func.name)
            elif isinstance(self._node, CStructField):
                name_box.setText(self._node.field)

            name_box.selectAll()
        self._name_box = name_box

        label_layout = QHBoxLayout()
        label_layout.addWidget(name_label)
        label_layout.addWidget(name_box)
        self.main_layout.addLayout(label_layout)

        # suggestions
        suggest_label = QLabel(self)
        suggest_label.setText("Suggestions")

        suggestion_box = QListWidget()
        self._suggestion_box = suggestion_box
        suggestion_layout = QHBoxLayout()
        suggestion_layout.addWidget(suggest_label)
        suggestion_layout.addWidget(suggestion_box)
        self.main_layout.addLayout(suggestion_layout)

        # status label
        status_label = QLabel(self)
        self.main_layout.addWidget(status_label)
        self._status_label = status_label

        # buttons
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self._ok_button = buttons.button(QDialogButtonBox.Ok)
        self._ok_button.setEnabled(False)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_name_changed(self, new_text):  # pylint:disable=unused-argument

        if self._name_box is None:
            # initialization is not done yet
            return

        if self._name_box.name is None:
            # the variable name is invalid
            self._status_label.setText('Invalid')
            self._status_label.setProperty('class', 'status_invalid')
            self._ok_button.setEnabled(False)
        else:
            self._status_label.setText('Valid')
            self._status_label.setProperty('class', 'status_valid')
            self._ok_button.setEnabled(True)

        self._status_label.style().unpolish(self._status_label)
        self._status_label.style().polish(self._status_label)

    def _on_ok_clicked(self):
        node_name = self._name_box.name
        if node_name is not None:
            if self._code_view is not None and self._node is not None:
                # need workspace for altering callbacks of changes
                workspace = self._code_view.instance.workspace
                code_kb = self._code_view.codegen.kb

                # stack variable
                if isinstance(self._node, CVariable) and self._node.unified_variable is not None:
                    # sanity check that we are a stack var
                    if hasattr(self._node.variable, 'offset') and self._node.variable.offset is not None:
                        workspace.plugins.handle_stack_var_renamed(
                            code_kb.functions[self._node.variable.region],
                            self._node.variable.offset,
                            self._node.variable.name,
                            node_name,
                        )

                    self._node.unified_variable.name = node_name
                    self._node.unified_variable.renamed = True

                # global variable
                elif isinstance(self._node, CVariable) and not self._node.variable.region:
                    workspace.plugins.handle_global_var_renamed(
                        self._node.variable.addr,
                        self._node.variable.name,
                        node_name
                    )

                    self._code_view.instance.kb.labels[self._node.variable.addr] = node_name
                    self._node.variable.name = node_name
                    self._node.variable.renamed = True

                # function arg
                elif isinstance(self._node, CVariable):
                    workspace.plugins.handle_func_arg_renamed(
                        code_kb.functions[self._node.codegen.cfunc.addr],
                        0,
                        self._node.variable.name,
                        node_name
                    )

                    self._node.variable.name = node_name
                    self._node.variable.renamed = True

                # function name
                elif isinstance(self._node, CFunction):
                    workspace.plugins.handle_function_renamed(
                        code_kb.functions[self._node.codegen.cfunc.addr],
                        self._node.name,
                        node_name
                    )

                    code_kb.functions.get_by_addr(self._node.addr).name = node_name
                    self._node.name = node_name
                    self._node.demangled_name = node_name

                # function renaming (as a call)
                elif isinstance(self._node, CFunctionCall):
                    if self._node.callee_func is not None:
                        workspace.plugins.handle_function_renamed(
                            code_kb.functions[self._node.codegen.cfunc.addr],
                            self._node.name,
                            node_name
                        )

                        self._node.callee_func.name = node_name

                # struct renaming
                elif isinstance(self._node, CStructField):
                    # TODO add a better callback
                    # TODO prevent name duplication. reuse logic from CTypeEditor?
                    # TODO if this is a temporary struct, make it permanent and add it to kb.types
                    fields = [(node_name if n == self._node.field else n, t)
                              for n, t in self._node.struct_type.fields.items()]
                    self._node.struct_type.fields = OrderedDict(fields)

                    workspace.plugins.handle_struct_changed(
                        self._node.field,
                        node_name
                    )

                    self._node.field = node_name

                elif isinstance(self._node, SimType):
                    ty = self._node
                    ref = None
                    while True:
                        if isinstance(ty, TypeRef):
                            ref = ty
                            ty = ty.type
                        elif isinstance(ty, SimTypePointer):
                            ref = None
                            ty = ty.pts_to
                        else:
                            break
                    if isinstance(ty, NamedTypeMixin):
                        ty.name = node_name
                        if ref is not None:  # update the typeref as well
                            ref._name = node_name
                    else:
                        ty.label = node_name

                self._code_view.codegen.am_event()
                self.close()



# pylint:disable=missing-class-docstring
from typing import TYPE_CHECKING
import json
import threading
import datetime
from time import sleep

from sqlalchemy import func as sqlalchemy_func
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QTableWidget, QHeaderView, \
    QAbstractItemView, QTableWidgetItem, QWidget, QTabWidget, QLabel

from angrmanagement.ui.views.view import BaseView
from angrmanagement.logic.threads import gui_thread_schedule_async

try:
    from slacrs import Slacrs
    from slacrs.model import PluginMessage
except ImportError as ex:
    Slacrs = None

if TYPE_CHECKING:
    from .chess_connector import ChessConnector


#
# Plugin Log
#

class QPluginItem:
    def __init__(self, plugin, kind, image_id, msg):
        self.plugin = plugin
        self.kind = kind
        self.image_id = image_id
        self.msg = msg

    def widgets(self):
        widgets = [
            QTableWidgetItem(self.plugin),
            QTableWidgetItem(self.kind),
            QTableWidgetItem(self.image_id),
            QTableWidgetItem(self.msg),
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QPluginsLogTable(QTableWidget):
    HEADER = [
        'Plugin',
        'Kind',
        'Image ID',
        'Message',
    ]

    def __init__(self, workspace, parent=None):
        super().__init__(parent)
        self.workspace = workspace

        self.setColumnCount(len(self.HEADER))
        self.setHorizontalHeaderLabels(self.HEADER)
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().setVisible(False)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)
        self.verticalHeader().setDefaultSectionSize(24)

        self.items = []
        self.cnt = 0

    def reload(self):
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()

    def update_table(self, messages):
        self.items = []

        for msg in messages:
            self.items.append(
                QPluginItem(
                    msg.plugin,
                    msg.kind,
                    msg.target_image_id,
                    msg.message
                )
            )

        self.reload()


#
#
#

class QFuzztainerItem:
    def __init__(self, data):
        self.data = data

    def widgets(self):
        widgets = [
            QTableWidgetItem(self.data),
        ]

        for w in widgets:
            w.setFlags(w.flags() & ~Qt.ItemIsEditable)

        return widgets


class QFuzztainerTable(QTableWidget):
    DISPLAY_ORDER = [
        'execs_per_sec',
        'unique_crashes',
        'unique_hangs',
        'paths_found',
        'paths_total',
        'testcache_size',
        'testcache_count',
        'testcache_evict',
        'run_time',
        'cycles_done',
        'cycles_wo_finds',
        'execs_done',
        'paths_favored',
        'max_depth',
        'pending_favs',
        'pending_total',
        'variable_paths',
        'execs_since_crash',
        'slowest_exec_ms',
        'edges_found',
    ]

    HEADER = [
        'execs/sec',
        'unique crashes',
        'unique hangs',
        'paths found',
        'paths total',
        'testcache size',
        'testcache count',
        'testcache evict',
        'run time',
        'cycles done',
        'cycles w/o finds',
        'execs done',
        'paths favored',
        'max depth',
        'pending favs',
        'pending total',
        'variable paths',
        'execs since crash',
        'slowest exec ms',
        'edges found',
    ]

    def __init__(self, workspace, idx, parent=None):
        super().__init__(parent)
        self.workspace = workspace
        self.idx = idx

        self.setRowCount(len(self.HEADER[self.idx*10:self.idx*10 + 10]))
        self.setVerticalHeaderLabels(self.HEADER[self.idx*10:self.idx*10 + 10])
        self.setColumnCount(1)

        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setVisible(False)

        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.SingleSelection)

        self.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
        self.verticalHeader().setSectionResizeMode(QHeaderView.Stretch)
        #self.verticalHeader().setSectionResizeMode(QHeaderView.Fixed)

        self.items = []

    def reload(self):
        self.setRowCount(len(self.items))

        for idx, item in enumerate(self.items):
            for i, it in enumerate(item.widgets()):
                self.setItem(idx, i, it)

        self.viewport().update()

    def update_table(self, messages):
        if messages is None:
            return

        self.items = []

        messages = messages[::-1]
        for msg in messages:
            if msg.plugin.lower() == "fuzztainer" and msg.kind == "info":
                try:
                    stats = json.loads(msg.message)
                except Exception as ex:  # pylint:disable=broad-except
                    print(ex)
                    continue

                try:
                    stats['execs_per_sec']
                except KeyError as e:
                    print(e)
                    continue

                for stat_name in self.DISPLAY_ORDER[self.idx*10:self.idx*10 + 10]:
                    stat_data = stats[stat_name]
                    self.items.append(
                        QFuzztainerItem(stat_data)
                    )

                break

        self.reload()


class SummaryView(BaseView):
    def __init__(self, workspace, default_docking_position, connector: 'ChessConnector', *args, **kwargs):
        super().__init__("chess_summary", workspace, default_docking_position, *args, **kwargs)
        self.base_caption = "CHESS"

        self.workspace = workspace
        self.connector = connector

        self._init_widgets()

        # start the worker routine
        self.should_work = Slacrs is not None
        self.worker_thread = threading.Thread(target=self.worker_routine)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()

    def teardown(self):
        self.should_work = False
        self.close()

    #
    # UI
    #

    def _init_widgets(self):
        self.fuzztainer_table_1 = QFuzztainerTable(self.workspace, 0)
        self.fuzztainer_table_2 = QFuzztainerTable(self.workspace, 1)
        self.log_table = QPluginsLogTable(self.workspace)

        #
        # fuzztainer tab
        #

        self.fuzztainer_tab = QWidget()
        fuzz_layout = QHBoxLayout()
        fuzz_layout.addWidget(self.fuzztainer_table_1)
        fuzz_layout.addWidget(self.fuzztainer_table_2)
        self.fuzztainer_tab.setLayout(fuzz_layout)

        #
        # plugin log tab
        #

        self.log_tab = QWidget()
        log_layout = QHBoxLayout()
        log_layout.addWidget(self.log_table)
        self.log_tab.setLayout(log_layout)

        #
        # final tab setup
        #

        # tabs for multiple summaries
        self.tabView = QTabWidget()
        self.tabView.addTab(self.fuzztainer_tab, "Fuzztainer")
        self.tabView.addTab(self.log_tab, "Plugins Log")

        layout = QVBoxLayout()
        layout.addWidget(self.tabView)

        #
        # status bar
        #
        self.status_bar = QLabel()

        layout.addWidget(self.status_bar)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.show()

    def set_status(self, text: str):
        self.status_bar.setText(text)

    #
    # Data Threading
    #

    def worker_routine(self):
        while self.should_work:
            if self.connector is None or not self.connector.connected:
                gui_thread_schedule_async(self.set_status, args=('Not connected to CHECRS backend.',))
                sleep(10)
                continue

            if not self.connector.target_image_id:
                gui_thread_schedule_async(self.set_status, args=('No associated remote target.',))
                sleep(2)
                continue

            sleep(5)

            gui_thread_schedule_async(self.set_status, args=('Refreshing...', ))
            session = None
            try:
                session = self._init_session()

                if not session:
                    continue

                self._update_tables(session, self.connector.target_image_id)
            finally:
                if session is not None:
                    session.close()

            last_updated = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            gui_thread_schedule_async(self.set_status, args=(f'Ready. Last updated at {last_updated}.',))

    def _init_session(self):
        if self.connector is None:
            self.connector = self.workspace.plugins.get_plugin_instance_by_name("ChessConnector")
            if self.connector is None:
                return None

        try:
            slacrs_instance = self.connector.slacrs_instance()
        except Exception:
            return None

        if slacrs_instance is None:
            return None

        session = slacrs_instance.session()
        return session

    @staticmethod
    def _find_latest_fuzztainer_stats(messages):
        messages = messages[::-1]
        for msg in messages:
            if msg.plugin.lower() == "fuzztainer" and msg.kind == "info":
                try:
                    stats = json.loads(msg.message)
                except Exception as e:
                    print(e)
                    continue

                try:
                    stats['execs_per_sec']
                except KeyError as e:
                    print(e)
                    continue

                return [msg]
        return None

    def _update_tables(self, session, target_image_id: str):
        res = session.query(PluginMessage).filter(
            sqlalchemy_func.lower(PluginMessage.plugin)=="fuzztainer",
            PluginMessage.kind=="info",
            PluginMessage.target_image_id==target_image_id,
        ).order_by(
            PluginMessage.created_at.desc()
        ).limit(50)
        if not res:
            return

        messages = res.all()
        if not messages:
            return

        fuzztainer_msgs = self._find_latest_fuzztainer_stats(messages)
        self.log_table.update_table(messages)
        self.fuzztainer_table_1.update_table(fuzztainer_msgs)
        self.fuzztainer_table_2.update_table(fuzztainer_msgs)



from typing import Optional, Tuple, TYPE_CHECKING
from PySide6.QtCore import Qt
from enum import Enum

if TYPE_CHECKING:
    from angrmanagement.ui.widgets.qblock import QBlock
    from angrmanagement.ui.widgets.qoperand import QOperand
    from angrmanagement.ui.views import DisassemblyView


class DisassemblyLevel(Enum):
    MachineCode = 0
    LifterIR = 1
    AIL = 2


class QDisassemblyBaseControl:
    """
    The base control class of QLinearViewer and QDisassemblyGraph. Implements or declares common shorthands and methods.
    """

    def __init__(self, instance, disasm_view, base_cls):
        self.instance = instance
        self.disasm_view: 'DisassemblyView' = disasm_view
        self._base_cls = base_cls
        self._insaddr_to_block = { }
        self._disassembly_level = disasm_view.disassembly_level

    @property
    def infodock(self):
        return self.disasm_view.infodock

    def refresh(self):
        """
        Recalculate sizes and positions of subitems, and trigger a refresh on every subitem (in order to reload text,
        etc.)

        :return:    None
        """
        raise NotImplementedError()

    def reload(self):
        raise NotImplementedError()

    def show_instruction(self, insn_addr, insn_pos=None, centering=False, use_block_pos=False, use_animation=False):
        raise NotImplementedError()

    #
    # Public methods
    #

    def get_selected_operand_info(self) -> Optional[Tuple['QBlock',int,'QOperand']]:
        if not self.infodock.selected_operands:
            return None

        # get the first operand
        ins_addr, operand_idx = next(iter(self.infodock.selected_operands))
        block = self._insaddr_to_block.get(ins_addr, None)
        if block is not None:
            operand = block.addr_to_insns[ins_addr].get_operand(operand_idx)
            return block, ins_addr, operand

        return None

    def set_disassembly_level(self, level:DisassemblyLevel):
        self._disassembly_level = level
        self.reload()

    #
    # Event handlers
    #

    def keyPressEvent(self, event):

        key = event.key()

        if key == Qt.Key_N:
            # rename a label
            self.disasm_view.popup_rename_label_dialog()
            return
        elif key == Qt.Key_X:
            # XRef
            # try to get a selected operand
            r = self.get_selected_operand_info()
            if r is not None:
                # xref to an operand
                _, ins_addr, operand = r
                self.disasm_view.parse_operand_and_popup_xref_dialog(ins_addr, operand)
                return

            # try to get a selected label
            if len(self.infodock.selected_labels) == 1:
                lbl_addr = next(iter(self.infodock.selected_labels))
                self.disasm_view.popup_xref_dialog(addr=lbl_addr, dst_addr=lbl_addr)
                return

        self._base_cls.keyPressEvent(self, event)





class FunctionHeader:

    __slots__ = ('name', 'prototype', 'args', )

    def __init__(self, name, prototype=None, args=None):
        self.name = name
        self.prototype = prototype
        self.args = args


class Variables:

    __slots__ = ['variables']

    def __init__(self, variables):
        self.variables = variables


class PhiVariable(Variables):

    __slots__ = ['variable']

    def __init__(self, variable, variables):
        super().__init__(variables)
        self.variable = variable


class Label:

    __slots__ = ['addr', 'text']

    def __init__(self, addr, text):
        self.addr = addr
        self.text = text
